/** \file hstlib.c
 * \brief TINE Local History Server module plus API routines
 *
 * A TINE server can be instructed to maintain a local history of certain properties.  The
 * stored data can then be obtained with normal data acquition calls to the TINE Local History
 * Server.  By and large the instructions as to which properties are to be maintained in the
 * local history server and as to what filtering criteria are to be used in commiting the data
 * to disk are typically entered into a startup configuration file 'history.csv'.  As an 
 * alternative, see AppendHistoryInformation().  The local history server maintains two sets
 * of histories, a short-term history kept in main memory and a long-term history kept on disk.
 * Whereas the short-term history can be kept on a millisecond time raster, the long-term history
 * will not be written to disk any more often than once per second.
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "srvcore.h"
#include "errors.h"
#include "swaplib.h"
#include "dbtools.h"
#include "toolkit.h"
#include "hstlib.h"
#ifdef WIN32
# include "direct.h"
#endif
#ifdef UNIX
# include <sys/stat.h>
# include <dirent.h>
# include <sys/vfs.h>
# ifndef FREEBSD
# include <alloca.h>
# endif
#endif
#ifdef VMS
# include <file.h>
# include <unixio.h>
#endif

#if defined(DOS)
# define O_createmode O_RDWR|O_CREAT|O_APPEND|O_BINARY
# define O_writemode O_RDWR|O_BINARY
# define O_readmode O_RDONLY|O_BINARY
# define O_mode O_RDONLY|O_BINARY
# define FO_mode "rt"
#elif defined(WIN32)
# define O_createmode O_RDWR|O_CREAT|O_APPEND|O_BINARY,_S_IREAD|_S_IWRITE
# define O_writemode O_RDWR|O_BINARY,_S_IREAD|_S_IWRITE
# define O_readmode O_RDONLY|O_BINARY
# define O_mode O_RDONLY|O_BINARY
# define FO_mode "rt"
#elif defined(VMS)
# define O_createmode O_RDWR|O_CREAT|O_APPEND
# define O_writemode O_RDWR
# define O_readmode O_RDONLY
# define O_mode O_RDONLY
# define FO_mode "r"
#else
# define O_createmode O_RDWR|O_CREAT|O_APPEND,SH_DENYNONE
# define O_writemode O_RDWR,SH_DENYNONE
# define O_readmode O_RDONLY,SH_DENYNONE
# define O_mode O_RDONLY
# define FO_mode "r"
#endif

extern char FecDBpath[80];
char ArchiveLocation[HISTORY_HOME_SIZE];
HstTblEntry *hstTbl = NULL;
int nHistoryRecords = 0;
TINE_EXPORT int maxHistoryRecords = MAX_HISTORY_RECORDS;
TINE_EXPORT int minHistoryPollingRate = 1000;
TINE_EXPORT int useMonthlyHistoryFiles = FALSE;
TINE_EXPORT int useHighResolutionTimeStamp = FALSE;
int gIsUsingStdFileSet = FALSE;
#define SEC_IN_DAY (3600L * 24L)
#define AS_STAT_SIZE 16

#define MakeScanFileName(nm,id) (\
  sprintf((nm),"%s*.%0x",GetFilePrefix(id),hstTbl[(id)].recordIndex)\
)

#define MakeTargetFileName(nm,yy,mm,dd,id) (\
  sprintf((nm),"%s%02d%02d%02d.%0x",GetFilePrefix(id),((yy)%100),(mm),\
  useMonthlyHistoryFiles ? 1 : (dd),hstTbl[(id)].recordIndex)\
)

int nDaysInMonthTable[2][12] = {
  {31,28,31,30,31,30,31,31,30,31,30,31},
  {31,29,31,30,31,30,31,31,30,31,30,31},
};
#define nDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonthTable[0][mm] : nDaysInMonthTable[1][mm])

int populateHRS(char *fecname,DBLNKLIST *fecList,HRIS **hris);

TINE_EXPORT void SetUseStandardHistoryFiles(int value)
{
# ifndef VXWORKS
  gIsUsingStdFileSet = value;
# else
  printf("SetUseStandardHistoryFiles has no effect for %s",OS_TAG);
# endif
}
char *GetFilePrefix(int id)
{
  if (useHighResolutionTimeStamp)
  {
    return hstTbl[(id)].useStandardSet ? "hs" : "ha";
  }
  else
  {
    return hstTbl[(id)].useStandardSet ? "as" : "ar";
  }
}
static UINT32 fsBlocksFree = 0;
UINT32 getAvailableDiskSpaceInBlocks(void)
{
# if defined(UNDER_CE)
  ULARGE_INTEGER freeBytesCaller;
  ULARGE_INTEGER totalBytes;
  ULARGE_INTEGER totalFreeBytes;
  
  if (GetDiskFreeSpaceEx(ArchiveLocation,&freeBytesCaller,&totalBytes,
      &totalFreeBytes) == 0) return 0;
  return (UINT32)freeBytesCaller.QuadPart;
# elif defined(WIN32)
  UINT32 SectorsPerCluster;
  UINT32 BytesPerSector;
  UINT32 NumberOfFreeClusters;
  UINT32 TotalNumberOfClusters;

  if (GetDiskFreeSpace(ArchiveLocation,&SectorsPerCluster,&BytesPerSector,
      &NumberOfFreeClusters,&TotalNumberOfClusters) == 0) return 0;
  return (NumberOfFreeClusters / 1000) * SectorsPerCluster * BytesPerSector;
# elif defined(UNIX)
# ifdef SOLARIS
  struct statvfs fsbuf;
  if (statvfs(ArchiveLocation, &fsbuf) != 0) return 0;
# else
  struct statfs fsbuf;
  if (statfs(ArchiveLocation, &fsbuf) != 0) return 0;
# endif
  return fsbuf.f_bavail;
# else
  return 0;
# endif
}

int getWorstCaseNumRecordsInFile(HstTblEntry *hst)
{
  double ndps;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(UINT32);
  if (hst == NULL) return -argument_list_error;
  ndps = 1000.0 / (double)hst->archiveRate;
  ndps *= SEC_IN_DAY;
  if (useMonthlyHistoryFiles)
  { /* how many times per month is possible ? */
    return (int)ndps * 31;
  }
  /* how many times per day is possible ? */
  return (int)ndps;
}
int getWorstCaseFileSpaceSizeBlocks(HstTblEntry *hst)
{
  int space = 0,recsPerFile,recSize;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(UINT32);

  if (hst->depthLong < 0) return 0;
  recsPerFile = getWorstCaseNumRecordsInFile(hst);
  recSize = hst->recordLength + timestampLength;
  space = recsPerFile * recSize + AS_STAT_SIZE;
  space /= 1000; /* blocks */
  space *= (hst->depthLong+1);
  return space;
}
#define MAX_NUM_OUTDATEDFILES 1000
NAME32 *outdatedFiles = NULL;
int nOutdatedFiles = 0;
int scanForOutdatedFiles(int index,int mdepth)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int n = 0, mm, yy, fmm, fyy;
  char dirname[128],filename[128],fname[64],fileext[16],fn[64],str[128];
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
  long hfp;
# endif
  struct tm *lt;
  time_t now = time(NULL);

  if (outdatedFiles == NULL) return -argument_list_error;
  if (index < 0 || index >= nHistoryRecords) return -invalid_index;
  /* what year and month is it now ? */
  lt = localtime(&now); 
  mm = lt->tm_mon - mdepth + 1;
  yy = lt->tm_year % 100;
  while (mm < 1) 
  {
    mm += 12;
    yy -= 1;
  }

  MakeScanFileName(fname,index);
  sprintf(fileext,".%0x",hstTbl[index].recordIndex);

# if defined(UNIX)
  sprintf(dirname,"%s",ArchiveLocation);
  if ((dp=opendir(dirname)) == NULL) return 0;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strstr(dirp->d_name,fileext) == NULL) continue;
    strncpy(fn,dirp->d_name,32); fn[32] = 0;
    sprintf(filename,"%s%s",ArchiveLocation,dirp->d_name);
# elif defined(WIN32)
  sprintf(dirname,"%s%s",ArchiveLocation,fname);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (strstr(fd.name,fileext) == NULL) continue;
    strncpy(fn,fd.name,32); fn[32] = 0;
    sprintf(filename,"%s%s",ArchiveLocation,fd.name);
# endif
    strncpy(str,&fn[2],2); str[2] = 0;
    fyy = atoi(str);
    strncpy(str,&fn[4],2); str[2] = 0;
    fmm = atoi(str);
    if (fyy > yy) continue;
    if (fyy == yy && fmm >= mm) continue;
    if (n >= MAX_NUM_OUTDATEDFILES) continue;
    strncpy(outdatedFiles[n].name,fn,32);
    n++;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
# endif
  return n; 
# else
  index = index;
  mdepth = mdepth;
  return 0;
# endif /* NO_LTS */
}
int checkForStandardFiles(void)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  char dirname[128];
  int filesfound = FALSE;
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  struct _finddata_t fd;
  long hfp;
# endif

# if defined(UNIX)
  sprintf(dirname,"%s",ArchiveLocation);
  if ((dp=opendir(dirname)) == NULL) return FALSE;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strncmp(dirp->d_name,"hs",2) != 0 &&
        strncmp(dirp->d_name,"as",2) != 0 ) continue;
# elif defined(WIN32)
  sprintf(dirname,"%s*.*",ArchiveLocation);
  if ((hfp=_findfirst(dirname, &fd)) == -1) return 0;
  do
  {
    if (strncmp(fd.name,"hs",2) != 0 &&
        strncmp(fd.name,"as",2) != 0) continue;
# endif
    filesfound = TRUE;
    break;
# if defined(UNIX)
  }
  closedir(dp);
# elif defined(WIN32)
  } while (_findnext(hfp,&fd) != -1);
# endif
  return filesfound;
# else
  return FALSE;
# endif /* NO_LTS */
}
int resetStats(char *fname,int recsize)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int hFile = -1, hAltFile = -1, fpos = 0;
  UINT32 statbuf[16];
  char fileName[128], altFileName[128], aname[64];
  BYTE *tmpHistoryDataBuffer;
  time_t t = 0;

  if (fname == NULL || recsize == 0) return argument_list_error;

  sprintf(fileName,"%s%s",ArchiveLocation,fname);
  /* prepare alternate file name in order to re-use old-style data */
  strncpy(aname,fname,64);
  if (!strncmp(aname,"hs",2)) strncpy(aname,"ha",2);
  if (!strncmp(aname,"as",2)) strncpy(aname,"ar",2);
  if (aname[1] != 's')
  { /* this should always be true, but check anyway */
    sprintf(altFileName,"%s%s",ArchiveLocation,aname);
    hAltFile = open(altFileName,O_readmode);
  }
  if ((hFile=open(fileName,O_writemode)) == -1)
  {
    if (tineDebug) perror(fileName);
    return file_error;
  }
  if (hAltFile != -1)
  { /* there was an old guy -> re-use the data */
    lseek(hFile,AS_STAT_SIZE,SEEK_SET);
  # ifdef INTEL
    tmpHistoryDataBuffer = (BYTE *)alloca(recsize);
  # else
    tmpHistoryDataBuffer = tmpWorkArea;
  # endif
    lseek(hAltFile,0,SEEK_SET);
    while(read(hAltFile,(char *)tmpHistoryDataBuffer,recsize) == recsize)   /* read through file */
    {
      if (write(hFile,(char *)tmpHistoryDataBuffer,recsize) != recsize) break;
      fpos += recsize;
    }
    t = time(NULL); /* give it the current time */
    close(hAltFile);
    /* now remove the old file */
    unlink(altFileName);
  }
  statbuf[0] = fpos;
  statbuf[1] = (UINT32)t;
  statbuf[2] = recsize;
  statbuf[3] = 0;
  lseek(hFile,0L,SEEK_SET);
  write(hFile,(char *)statbuf,16);

  close(hFile);
# else
  fname = fname;
# endif
  return 0;
}
/**
 * \internal
 *
 * \brief create the standard archive files 
 * according to depth in months, record size and archive rate.
 * 
 */
int makeStandardFiles(int index,int mdepth,int mday)
{
# if (defined(UNIX) || defined(WIN32)) && !defined(NO_LTS)
  int hFile = -1, hStat = -1, mm, yy, ptr, isCurrentFile = FALSE, wasFound = FALSE;
  int i, cc = 0, recsPerFile, recSize, bytesWritten, needResetStats = FALSE;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(UINT32);
  char fileName[128], fname[64], tmp[128];
  UINT32 statbuf[4];
  BYTE *tmpHistoryDataBuffer;
  time_t t=time(NULL), ftime;
  struct tm *lt, *ft;
  HstTblEntry *hst;

  if (index < 0 || index >= nHistoryRecords) return invalid_index;
  hst = &hstTbl[index];

  lt = localtime(&t); 
  mm = lt->tm_mon - mdepth + 1;
  yy = lt->tm_year;

  if (mdepth == 0 && (useMonthlyHistoryFiles || mday == lt->tm_mday))
  {
    isCurrentFile = TRUE;
  }
  else
  {
    while (mm < 1) 
    {
      mm += 12;
      yy -= 1;
    }
  }

  MakeTargetFileName(fname,yy,mm,mday,index);
  sprintf(fileName,"%s%s",ArchiveLocation,fname);
  if (tineDebug) dbglog("make standard local history file : %s",fileName);

  if (nOutdatedFiles > 0)
  { /* there are outdated files -> re-use them ! */
    if ((hFile=open(fileName,O_mode)) == -1)
    { /* not there, so rotate an old one */
      feclog("recycle history file %s to %s",outdatedFiles[nOutdatedFiles].name,fname);
      nOutdatedFiles--;
      sprintf(tmp,"%s%s",ArchiveLocation,outdatedFiles[nOutdatedFiles].name);
      rename(tmp,fileName);
    }
    else
    { /* was there ! */
      close(hFile);
    }
  }

  if ((hFile=open(fileName,O_createmode)) == -1)
  {
    if (tineDebug) perror(fileName);
    return file_error;
  }

  recsPerFile = getWorstCaseNumRecordsInFile(hst);
  recSize = hst->recordLength + timestampLength;

  if ((ptr=lseek(hFile,0L,SEEK_END)) > 0) 
  { /* found a file; is it big enough ? */
    wasFound = TRUE;
    if (ptr >= recsPerFile * recSize + AS_STAT_SIZE) ccerr(0);
    recsPerFile -= (ptr-AS_STAT_SIZE)/recSize;  /* append the difference */
  }
  if (mday == 1)
  {
    feclog("creating file space for history record %d (depth %d months)",hst->recordIndex,mdepth);
    if (!useMonthlyHistoryFiles) feclog("file space includes 31 separate files per month");
    feclog("file space for history record %d requires %d blocks",hst->recordIndex,getWorstCaseFileSpaceSizeBlocks(hst));
  }
# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(recSize);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  /* create the 16 bytes of file status */
  memset(statbuf,0,AS_STAT_SIZE);
  /* create or append to the standard file */
  memset(tmpHistoryDataBuffer,0,recSize);
  if (!wasFound) write(hFile,(char *)statbuf,AS_STAT_SIZE);
  for (i=0; i<recsPerFile; i++)
  { /* file in the worst case standard file (minimal fragmentation) */
    if ((bytesWritten=write(hFile,(char *)tmpHistoryDataBuffer,recSize)) != recSize)
    {
      feclog("unable to allocate local history file space for %s",fileName);
      ccerr(no_file_space);
    }
  }
err:
  /* are there records written to the file ? */
  if (wasFound && cc == 0)
  { /* not the first startup !*/
    if (isCurrentFile)
    { /* ensure that file time is not too old */
      lseek(hFile,0L,SEEK_SET);
      if (read(hFile,(char *)&statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) 
      {
        feclog("error reading local history standard file : %s",fileName);
        return file_error;
      }
      ptr = statbuf[0];
      ftime = (time_t)statbuf[1];
      if (recSize != (int)statbuf[2]) 
      {
        feclog("needed to adjust record size in local history standard file : %s",fname);
        needResetStats = TRUE;
      }
      ft = localtime(&ftime); 
      if (ft->tm_year != yy || ft->tm_mon != mm - 1) needResetStats = TRUE;
      if (!useMonthlyHistoryFiles && ft->tm_mday != mday) needResetStats = TRUE;
    }
  }
  if (hFile != -1) 
  {
    close(hFile);
    if (!wasFound || needResetStats)
    {  /* fill in position = 0, UTC = 0 */
      if (cc == 0) resetStats(fname,recSize);
    }
  }

  return cc;
# else
  index = index;
  mdepth = mdepth;
  mday = mday;
  return 0;
# endif /* NO_LTS */
}
int rotateHistoryFile(char *srcfn,char *dstfn)
{
# if !defined(NO_LTS) && !defined(VXWORKS)
  int cc = 0;
  char srcfile[128],dstfile[128];
  struct stat sbuf;

  if (srcfn == NULL || dstfn == NULL) ccerr(argument_list_error);

  sprintf(dstfile,"%s%s",ArchiveLocation,dstfn);
  if (stat(dstfile,&sbuf) == 0) ccerr(0); /* already there */
  sprintf(srcfile,"%s%s",ArchiveLocation,srcfn);
  if (rename(srcfile,dstfile) != 0) ccerr(file_error);
err:
  if (cc) feclog("rotateHistoryFile %s -> %s : %s",srcfn,dstfn,erlst[cc]);
  return cc;
# else
  srcfn = srcfn;
  dstfn = dstfn;
  return 0;
# endif
}
int getFilePosition(int hFile,int useStdFileSet)
{
  int cc = 0, pos;
  UINT32 statbuf[4];
  if (useStdFileSet)
  {
    lseek(hFile,0L,SEEK_SET);
    if (read(hFile,(char *)&statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) ccerr(file_error);
    pos = (int)statbuf[0] + AS_STAT_SIZE;
  }
  else
  {
    pos = lseek(hFile,0L,SEEK_END);
  }
err:
  return cc != 0 ? -cc : pos;
}
int setFilePosition(int hFile,int pos)
{
  int cc = 0;
  time_t t = time(NULL);
  UINT32 statbuf[4];
  lseek(hFile,0L,SEEK_SET);
  statbuf[0] = pos - AS_STAT_SIZE;
  statbuf[1] = (UINT32)t;
  if (write(hFile,(char *)&statbuf,AS_STAT_SIZE/2) != AS_STAT_SIZE/2) ccerr(file_error);
err:
  return cc;
}
int getNumberOfRecords(int hFile,int recsize,int useStdFileSet)
{
  int cc = 0, pos, siz;
  UINT32 statbuf[4];
  if (useStdFileSet)
  {
    lseek(hFile,0L,SEEK_SET);
    if (read(hFile,(char *)&statbuf,AS_STAT_SIZE) != AS_STAT_SIZE) ccerr(file_error);
    pos = (int)statbuf[0];
    siz = (int)statbuf[2];
    if (siz != recsize) ccerr(invalid_data);
  }
  else
  {
    pos = lseek(hFile,0,SEEK_END);
    siz = recsize;
  }
err:
  return cc != 0 ? -cc : pos/siz;
}
int removeOutdatedLTSFile(char *fn)
{
  int filesRemoved = 0;
# ifndef NO_LTS
  int i;
  char fileName[128],lfn[32];
  static int stagger = -1;

  if (stagger < 0) stagger = (int)time(NULL); /* seed the start up */
  /* stagger file removal to avoid blocking the file system */
  stagger = (stagger + 1) % 100;
  if (stagger > 0) return -1; 

  for (i=1; i<32; i++)
  {
    strncpy(lfn,fn,6);
    sprintf(&lfn[6],"%02d",i);
    strcat(lfn,&fn[8]);
    sprintf(fileName,"%s%s",ArchiveLocation,lfn);
    if (unlink(fileName) == 0) filesRemoved++;
  }
# endif /* NO_LTS */
  if (filesRemoved > 0) feclog("HST: removed %d files from long term storage",filesRemoved);
  return filesRemoved;
}
int WriteRecordToLTS(int i)  /* write record #i to file */
{
# ifndef NO_LTS
  int hFile,ptr,fpos;
  HstTblEntry *hst;
  char fileName[128];
  BYTE *tmpHistoryDataBuffer;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(UINT32);
  SINT32 lTimeStamp;

  if (i < 0 || i >= nHistoryRecords) return invalid_index;
  hst = &hstTbl[i];
  ptr = hst->dataPtr;
  sprintf(fileName,"%s%s",ArchiveLocation,hst->ArchiveFileName);

  if ((hFile=open(fileName,hst->useStandardSet ? O_writemode : O_createmode)) == -1)
  {
    if (tineDebug) perror(fileName);
    return file_error;
  }
  if ((fpos=getFilePosition(hFile,hst->useStandardSet)) < 0) return file_error;
  lseek(hFile,fpos,SEEK_SET);

  /* write the timestamp and record as one unit to avoid corruption */
# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(hst->recordLength+timestampLength);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  if (useHighResolutionTimeStamp)
  {
    memcpy(tmpHistoryDataBuffer,(char *)&hst->timestamp[ptr],sizeof(double));
  }
  else
  {
    lTimeStamp = (SINT32)hst->timestamp[ptr];
    memcpy(tmpHistoryDataBuffer,(char *)&lTimeStamp,sizeof(UINT32));
  }
  memcpy(&tmpHistoryDataBuffer[timestampLength],hst->data + ptr * hst->recordLength,hst->recordLength);

  write(hFile,(char *)tmpHistoryDataBuffer,hst->recordLength+timestampLength);
  
  if (hst->useStandardSet)
  {
    fpos += hst->recordLength+timestampLength;
    setFilePosition(hFile,fpos);
  }
  close(hFile);

  hst->lastArchiveTime = hst->timestamp[ptr];
# endif /* NO_LTS */
  return 0;
}
SINT32 locateRecordFromLTS(char *fileName,int i,double dt)
{ /* p0,d0 < p1,d1 < p2,d2 */
# ifndef NO_LTS
  SINT32 t,t0,t1,t2;
  double dt0, dt1, dt2;
  SINT32 nrecs,p0,p1,p2;
  double f;
  int hFile,recsize,it=0;
  int lts_offset;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
  HstTblEntry *hst;

  if (i < 0 || i >= nHistoryRecords) return -invalid_index;
  hst = &hstTbl[i];
  recsize = hst->recordLength + timestampLength;

  if ((hFile=open(fileName,O_readmode)) == -1) return -file_error;
  t = (SINT32)dt;
  nrecs = getNumberOfRecords(hFile,recsize,hst->useStandardSet);
  lts_offset = hst->useStandardSet ? AS_STAT_SIZE : 0;
  p0 = p1 = lts_offset; p2 = (nrecs-1)*recsize + lts_offset;
  lseek(hFile,p0,SEEK_SET);
  if (useHighResolutionTimeStamp)
  {
    read(hFile,(char *)&dt0,timestampLength);
    t0 = (SINT32)dt0;
  }
  else
  {
    read(hFile,(char *)&t0,timestampLength);
    dt0 = (double)t0;
  }
  if (dt < dt0) goto out;
  lseek(hFile,p2,SEEK_SET); 
  if (useHighResolutionTimeStamp)
  {
    read(hFile,(char *)&dt2,timestampLength);
    t2 = (SINT32)dt2;
  }
  else
  {
    read(hFile,(char *)&t2,timestampLength);
    dt2 = (double)t2;
  }
  while (p2 > p0 + recsize)
  {
    if (dt > (dt1=dt2)) { p1 = p2; break; }
    while (dt1 > dt0)   /* seek <--  */
    {
      f = ((dt-dt0)/(dt1-dt0))*((double)(p2-p0)/recsize);
      if ((p1=p0+recsize*(SINT32)f) >= p2) break;
	    lseek(hFile,p1,SEEK_SET);
      if (useHighResolutionTimeStamp)
      {
        read(hFile,(char *)&dt1,timestampLength);
        t1 = (SINT32)dt1;
      }
      else
      {
        read(hFile,(char *)&t1,timestampLength);
        dt1 = (double)t1;
      }
      if (dt1 < dt) break;
      p2 = p1;
      t2 = t1;
      dt2 = dt1;
      it++;
    }
    dt0 = dt1; t0 = t1; p0 = p1;
    while (dt2 > dt1)  /* seek --> */
    {
      f = ((dt2-dt)/(dt2-dt1))*((double)(p2-p0)/recsize);
      if ((p1=p2-recsize*(SINT32)f) <= p0) break;
	    lseek(hFile,p1,SEEK_SET); 
      if (useHighResolutionTimeStamp)
      {
        read(hFile,(char *)&dt1,timestampLength);
        t1 = (SINT32)dt1;
      }
      else
      {
        read(hFile,(char *)&t1,timestampLength);
        dt1 = (double)t1;
      }
      if (dt1 > dt) break;
      p0 = p1; 
      t0 = t1;
      dt0 = dt1;
      it++;
    }
    p2 = p1; t2 = t1; dt2 = dt1;
  }
out: 
  close(hFile);
  return p1;
# else
  return 0;
# endif /* NO_LTS */
}
int needToArchiveRecord(int i)
{
  double t, ts = makeDataTimeStamp();
  BYTE *d0,*d1;
  HstTblEntry *hst;
  int p,ptr,delta_t,doarchive = 0;

  if (i < 0 || i >= nHistoryRecords) return FALSE;
  hst = &hstTbl[i];
  if (hst->archiveRate < 0 ||
      hst->heartbeat < 0   ||
      hst->depthLong < 0) return FALSE;/* no long term storage */
  ptr = hst->dataPtr;

  if (hst->timestamp[ptr] <= hst->lastArchiveTime) return FALSE;/* data timestamp hasn't changed */
  if (!useHighResolutionTimeStamp && (SINT32)hst->timestamp[ptr] == (SINT32)hst->lastArchiveTime) return FALSE;
  if (hst->lastArchiveTime == 0) hst->lastArchiveTime = ts - hst->archiveRate/1000.0;
  delta_t = (int)((ts - hst->lastArchiveTime)*1000);
  if (hst->isScheduled) delta_t = hst->archiveRate * 2;
  if (delta_t < hst->archiveRate) return FALSE;
  if (delta_t >= hst->heartbeat*1000) return TRUE; /* archive no matter what */

  d0 = hst->data + ptr * hst->recordLength;
  d1 = ptr > 0 ? hst->data + (ptr-1) * hst->recordLength :
                 hst->data + (hst->depthShort-1) * hst->recordLength;

  switch (LFMT(hst->c.EqmFormatOut))
  {
    case CF_DOUBLE:
         for (p=0; !doarchive && p<(int)hst->c.EqmSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * fabs(((double *)d1)[p]);
           if ( fabs(((double *)d0)[p] - ((double *)d1)[p]) > t) doarchive = 1;
         }
         break;
    case CF_FLOAT:
         for (p=0; !doarchive && p<(int)hst->c.EqmSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * fabs(((float *)d1)[p]);
           if ( fabs(((float *)d0)[p] - ((float *)d1)[p]) > t) doarchive = 1;
         }
         break;
    case CF_INTEGER:
         for (p=0; !doarchive && p<(int)hst->c.EqmSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * abs(((short *)d1)[p]);
           if ( abs(((short *)d0)[p] - ((short *)d1)[p]) > t) doarchive = 1;
         }
         break;
      case CF_LONG:
         for (p=0; !doarchive && p<(int)hst->c.EqmSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * labs(((SINT32 *)d1)[p]);
           if ( labs(((SINT32 *)d0)[p] - ((SINT32 *)d1)[p]) > t) doarchive = 1;
         }
         break;
      case CF_BYTE:
         for (p=0; !doarchive && p<(int)hst->c.EqmSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * abs(((BYTE *)d1)[p]);
           if ( abs(((BYTE *)d0)[p] - ((BYTE *)d1)[p]) > t) doarchive = 1;
         }
         break;
      default:
         if (!doarchive && memcmp(d0,d1,hst->recordLength)) doarchive = 1;
         break;
  }
  return doarchive ? TRUE : FALSE;
}
int getHistoryTableIndex(HstTblEntry *hst,int nentries)
{
  int i;
  for (i=0; i<nentries; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,hst->c.EqmName,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,hst->c.EqmProperty,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,hst->c.EqmDeviceName,DEVICE_NAME_SIZE)) continue;
    if (hstTbl[i].c.EqmSizeOut != hst->c.EqmSizeOut) continue;
    if (hstTbl[i].c.EqmFormatOut != hst->c.EqmFormatOut) continue;
    return i;
  }
  return -1;
}
int isInHistoryTable(HstTblEntry *hst,int nentries)
{
  if (getHistoryTableIndex(hst,nentries) < 0) return FALSE;
  feclog("HIST: ignoring double entry [%s] %s %s",hst->c.EqmName,hst->c.EqmDeviceName,hst->c.EqmProperty);
  return TRUE;
}
int makeFilesForRecord(int index)
{
  int i, cc, dd, dstart, dstop, dbglevel;
  HstTblEntry *hst;
  if (index < 0 || index >= nHistoryRecords) return invalid_index;
  hst = &hstTbl[index];
  dbglevel = tineDebug; /* note the current debug level */
  if (dbglevel < 1) tineDebug = 1; /* make sure it's at least turned on ! */
  /* make sure the standard (this month) file is prepared first */
  dstart = 1; 
  dstop = useMonthlyHistoryFiles ? 1 : 31;
  nOutdatedFiles = scanForOutdatedFiles(index,hst->depthLong);
  for (i=0; i<hst->depthLong+1; i++) 
  {
    for (dd=dstart; dd<dstop+1; dd++)
    {
      if ((cc=makeStandardFiles(index,i,dd)) != 0)
      {
        feclog("make local history files : %s",erlst[cc]);
      }
    }
  }
  tineDebug = dbglevel; /* set it back */
  return 0;
}

CSVCOLUMN csvHistoryInfoColumns[] = 
{
  { "INDEX","",0,CF_LONG,1,(BYTE *)OFFSETIN(HRIS,recordIndex),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,c.EqmName),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,c.EqmProperty),NULL,0 },
  { "DEVICE","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,c.EqmDeviceName),NULL,0 },
  { "DATA_LENGTH","1",0,CF_LONG,1,(BYTE *)OFFSETIN(HRIS,c.EqmSizeOut),NULL,0 },
  { "FORMAT","",0,CF_TEXT,64,(BYTE *)OFFSETIN(HRIS,fmt),NULL,0 },
  { "HEARTBEAT","900",0,CF_LONG,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(HRIS,heartbeat),NULL,0 },
  { "POLLING_RATE","1000",0,CF_LONG,1,(BYTE *)OFFSETIN(HRIS,pollingRate),NULL,0 },
  { "ARCHIVE_RATE","1000",0,CF_LONG,1,(BYTE *)OFFSETIN(HRIS,archiveRate),NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(HRIS,tolerance),NULL,0 },
  { "SHORT_DEPTH","0",0,CF_LONG,1,(BYTE *)OFFSETIN(HRIS,depthShort),NULL,0 },
  { "LONG_DEPTH","0",0,CF_LONG,1,(BYTE *)OFFSETIN(HRIS,depthLong),NULL,0 },
  { "EXPORT_NAME","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(HRIS,exportTarget),NULL,0 },
  { "REMOTE_SERVER","",0,CF_TEXT,96,(BYTE *)OFFSETIN(HRIS,redirection),NULL,0 }
};
CSVDB csvHistoryInfoDb =
{
  csvHistoryInfoColumns,
  sizeof(csvHistoryInfoColumns)/sizeof(csvHistoryInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(HRIS)
};

int LoadHistoryInformation(char *fname,char *expName,char *eqmName) /* read history record info from file */
{
  HRIS *hrs = NULL;
  int i, n, sz, cc = 0;
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);
  /* read in the details */
  if (isUsingXMLFecDatabase)
  {
    sz = populateHRS(gFecName,gFecInfoList,&hrs);
    if (sz < 0) ccerr(-sz);
  }
  else
  {
    if ((cc=csvReadFile(FecDBpath,fname,&csvHistoryInfoDb,(void **)&hrs)) != 0) ccerr(cc);
    sz = csvHistoryInfoDb.siz;
  }
  if ((n=GetNumberEquipmentModules()) == 0) n = 1;
  /* make adjustments to local table size */
  if (sz > maxHistoryRecords/2 && nHistoryRecords == 0) maxHistoryRecords = 2*sz*n;
  /* allocate space for local table (allow extra space for new entries) */
  if (hstTbl == NULL && (hstTbl=(HstTblEntry *)SystemCalloc(maxHistoryRecords,sizeof(HstTblEntry))) == NULL)
    ccerr(out_of_local_memory);
  for (i=nHistoryRecords; i<sz && i<maxHistoryRecords; i++)
  { /* prepare local history record table */
    if (eqmName != NULL && strncmp(hrs[i].c.EqmName,eqmName,EQM_NAME_SHORTSIZE)) continue;
    if (strlen(hrs[i].exportTarget) > 0 && expName != NULL)
    { /* legacy configuration trying to latch a registered export name */
      if (strnicmp(hrs[i].exportTarget,expName,EXPORT_NAME_SIZE)) continue; /* didn't match ! */
    }
    if (strchr(hrs[i].tolerance,'%')) 
    {
      hstTbl[i].pTolerance = (float)(atof(hrs[i].tolerance) / 100.0);
    }
    else
    {
      hstTbl[i].aTolerance = (float)atof(hrs[i].tolerance);
    }
    if (strlen(hrs[i].fmt) > 0) hrs[i].c.EqmFormatOut = BFMT(ftoi(hrs[i].fmt));
    memcpy(&hstTbl[i].c,&hrs[i].c,sizeof(CONTRACT));
    hstTbl[i].pollingRate = hrs[i].pollingRate;
    hstTbl[i].archiveRate = hrs[i].archiveRate;
    hstTbl[i].depthShort = hrs[i].depthShort;
    hstTbl[i].depthLong = hrs[i].depthLong;
    hstTbl[i].heartbeat = hrs[i].heartbeat;
    hstTbl[i].recordIndex = hrs[i].recordIndex;
    if (isInHistoryTable(&hstTbl[i],i)) continue;
    if (strlen(hrs[i].redirection) > 0)
    { /* property is redirected */
      if ((hstTbl[i].prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) 
        ccerr(out_of_server_memory);
      ParseRedirector(hrs[i].redirection,hstTbl[i].prd);
    }
    if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate *= 1000; /* was thinking in seconds ! */
    if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate = 1000; /* don't allow a zero ! */
    if (hstTbl[i].archiveRate < 20) hstTbl[i].archiveRate *= 1000; /* was thinking in seconds ! */
    if (hstTbl[i].archiveRate < 20) hstTbl[i].archiveRate = 1000; /* don't allow a zero ! */
    if (hstTbl[i].archiveRate < hstTbl[i].pollingRate) hstTbl[i].pollingRate = hstTbl[i].archiveRate;
    if (hstTbl[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = hstTbl[i].pollingRate;
    hstTbl[i].recordLength =
      (int)hstTbl[i].c.EqmSizeOut * fmtsizeof(hstTbl[i].c.EqmFormatOut);
    MakeTargetFileName(hstTbl[i].ArchiveFileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
    hstTbl[i].useStandardSet = gIsUsingStdFileSet;
    if ((hstTbl[i].timestamp=(double *)SystemCalloc(hstTbl[i].depthShort,sizeof(double))) == NULL)
      ccerr(out_of_local_memory);
    if ((hstTbl[i].data=(BYTE *)SystemCalloc(hstTbl[i].depthShort,hstTbl[i].recordLength)) == NULL)
      ccerr(out_of_local_memory);
    feclog("HIST: %s %s %s (%d s %d mon)",
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmProperty,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].depthShort,
            hstTbl[i].depthLong);
    nHistoryRecords++;
  }
  
  if (gIsUsingStdFileSet)
  { /* reading history file : true for one, true for all */
    feclog("using standard non-fragmented history files at initialization");
    for (i=0; i<nHistoryRecords; i++)
    {
      makeFilesForRecord(i);
    }
  }
  if (MaxPollingRate > minHistoryPollingRate)
  {
    SysPoll = minHistoryPollingRate;
    MaxPollingRate = minHistoryPollingRate;
    if (minHistoryPollingRate < MinPollingRate) MinPollingRate = minHistoryPollingRate;
  }
err:
  if (hrs != NULL) SystemFree(hrs);
  if (cc && cc != no_such_file)
  {
    feclog("HIST: %s : %s",fname,erlst[cc]);
  }
  return cc;
}

int history_home_done = FALSE;
void historyInit(char *expName,char *eqmName)
{
  static char lastExpName[EXPORT_NAME_SIZE] = "";
  static char lastEqmName[EXPORT_NAME_SIZE] = "";
  ExportListStruct *el;
  char *ptr, *c, hstfn[64];
  int i;

  if ((expName == NULL || !strnicmp(expName,lastExpName,EXPORT_NAME_SIZE)) &&
      (eqmName == NULL || !strncmp(eqmName,lastEqmName,EQM_NAME_SIZE)))
    return;
  if (expName == NULL) lastExpName[0] = 0; else strncpy(lastExpName,expName,EXPORT_NAME_SIZE);
  if (eqmName == NULL) lastEqmName[0] = 0; else strncpy(lastEqmName,eqmName,EQM_NAME_SIZE);
  if ((el=getExportListItem(eqmName)) != NULL && el->hstdone) return;
  feclog("HST: initializing local history subsystem for module %s",eqmName == NULL ? "(null)" : eqmName);
  if (!history_home_done)
  {
    if (strlen(ArchiveLocation) == 0)
    { /* could have been set from fec.xml or fecid.csv */
      if ((ptr=getenv("HISTORY_HOME")) == NULL && (ptr=getenv("HISTORYDB")) == NULL) 
      { /* not set from environment -> use FEC_HOME */
        ptr = FecDBpath;
      }
      if (*ptr == '\"') ptr++; 
      if ((c=strchr(ptr,'\"')) != NULL) *c = 0;
      strncpy(ArchiveLocation,ptr,80);
    }
  # ifdef FS_DELIMITER
    if (((i=(int)strlen(ArchiveLocation)) > 0) &&
        (ArchiveLocation[i-1] != FS_DELIMITER))  /* no delimiter ! */
    {
      ArchiveLocation[i] = FS_DELIMITER; ArchiveLocation[i+1] = 0;
    }
  # endif
    makeDirectoryPath(ArchiveLocation);
    feclog("HISTORY HOME : %s",ArchiveLocation);
    history_home_done = TRUE;
  }

  fsBlocksFree = getAvailableDiskSpaceInBlocks();
  feclog("file system has %d free blocks",fsBlocksFree);

  gIsUsingStdFileSet = checkForStandardFiles();
  if (gIsUsingStdFileSet) feclog("found standard non-fragmented history files in history home");

  outdatedFiles = (NAME32 *)SystemCalloc(MAX_NUM_OUTDATEDFILES,sizeof(NAME32));
  nOutdatedFiles = MAX_NUM_OUTDATEDFILES;

  strncpy(hstfn,HISTORY_FILENAME,64);
# ifdef FS_DELIMITER
  if (strlen(gFecDBSubDir) == 0 && eqmName != NULL)
  { /* in case of history.csv but no exports.csv file */
    strncpy(gFecDBSubDir,eqmName,16);
    gFecDBSubDir[strlen(gFecDBSubDir)] = FS_DELIMITER;
  }
# endif
  if (strlen(gFecDBSubDir) > 0) sprintf(hstfn,"%s%s",gFecDBSubDir,HISTORY_FILENAME);

  for (i=0; i<2; i++)
  {
    if (LoadHistoryInformation(hstfn,expName,eqmName) != no_such_file) break;
    strncpy(hstfn,HISTORY_FILENAME,64);
  }

  if (outdatedFiles != NULL) SystemFree(outdatedFiles); /* return memory to the heap */  
  outdatedFiles = NULL;
  nOutdatedFiles = 0;
  if (el != NULL) el->hstdone = TRUE;
}
static CLN hstcln;
static CONSUMER hstcon;
static ContractListStruct hstcl;
void historyCycle(int isScheduled)
{
  int i,n,p,cc;
  ExportListStruct *el;
  char sDevName[DEVICE_NAME_SIZE];
  struct tm *tm;
  int delta_t;
  static struct timeval lastCycleTime = {0, 0};
  static time_t lasttime = 0;
  static DTYPE din,dout;
  static int mday = -1;
  int yy,mm,dd,rmvrec,nstdfiles=0;
  char rmvfn[64];
  static int recordToRemove = 0;
  if (ServerExitCondition) return;
  if (!isScheduled)
  {
    if (MSECS(thisCycleTime,lastCycleTime) < (int)(minHistoryPollingRate-gSystemTick)) return;
    lastCycleTime = thisCycleTime;
  }
  if (lasttime == 0) 
  { /* initialization */
    strncpy(hstcln.userName,"LOCAL_HISTORY",USERNAME_SIZE);
    hstcon.client = &hstcln;
    hstcl.consumer = &hstcon;
    hstcl.nconsumers = 1;
    memset(&din,0,sizeof(DTYPE));
  }
  lasttime = thisCycleTime.tv_sec;
  tm = localtime(&lasttime);
  for (i=0; i<nHistoryRecords; i++)
  {
    if ((el=getExportListItem(hstTbl[i].c.EqmName)) == NULL) continue;
    if (el->eqmFcn == NULL) continue;
    if (el->eqmbusy) continue;
    if (hstTbl[i].depthShort <= 0) continue;
    if (mday != tm->tm_mday)
    { /* it's the dawning of a new day (sort of) */
      MakeTargetFileName(hstTbl[i].ArchiveFileName,tm->tm_year,tm->tm_mon+1,tm->tm_mday,i);
      rmvrec = hstTbl[i].useStandardSet ? i : recordToRemove;
      mm = ((tm->tm_mon+12) - ((hstTbl[rmvrec].depthLong+1)%12))%12;
      yy = tm->tm_year - (hstTbl[rmvrec].depthLong+1)/12;
      if (mm > tm->tm_mon) yy--;
      dd = mday < 0 ? 0 : mday;
      MakeTargetFileName(rmvfn,yy,mm+1,dd,rmvrec);
      if (!hstTbl[i].useStandardSet)
      { /* if not using a standard rotation set, then remove the old ones */
        if (recordToRemove >= i && recordToRemove < nHistoryRecords)
        { /* careful of module arithmetic here. uses months from 0 to 11 */
          if (removeOutdatedLTSFile(rmvfn) >= 0) recordToRemove++;
        }
      }
      else
      { /* rotate the outdated file into the current position */
        if (mday > 0)
        {
          rotateHistoryFile(rmvfn,hstTbl[i].ArchiveFileName);
        }
        nstdfiles++; /* count the standard files (not to be removed) */
      }
    }
    gDataTimeStamp = PutDataTimeStamp(gDataTimeStampOffset,thisCycleTime.tv_sec,(int)(thisCycleTime.tv_usec/1000));
    n = hstTbl[i].dataPtr;
    if (hstTbl[i].timestamp[n] == 0) 
      hstTbl[i].timestamp[n] = gDataTimeStamp - (double)hstTbl[i].pollingRate/1000.0;
    delta_t = (int)((gDataTimeStamp - hstTbl[i].timestamp[n])*1000);
    if (hstTbl[i].isScheduled) delta_t = (int)hstTbl[i].pollingRate * 2;
    if (delta_t < (int)(hstTbl[i].pollingRate-gSystemTick)) continue;
    n = (hstTbl[i].dataPtr + 1) % hstTbl[i].depthShort;
    dout.dArrayLength = hstTbl[i].c.EqmSizeOut;
    dout.dFormat = LFMT(hstTbl[i].c.EqmFormatOut);
    dout.data.vptr = hstTbl[i].data + n*hstTbl[i].recordLength;
    strncpy(sDevName,hstTbl[i].c.EqmDeviceName,DEVICE_NAME_SIZE);
    hstcl.exp = el;
    hstcl.compStatus = not_signalled;
    cc = callEqm(&hstcl,sDevName,hstTbl[i].c.EqmProperty,&dout,&din,CA_READ|CA_HIST);
    if (cc != 0)
    {
      if (tineDebug)
      {
        dbglog("HIST: (%d) %s %s : <%d> %s",
                i,sDevName,hstTbl[i].c.EqmProperty,
                cc,cc > 0 && cc < numErr ? erlst[cc] : "undefined");
      }
      hstTbl[i].isScheduled = FALSE;
      continue;
    }
    p = hstTbl[i].dataPtr;
    delta_t = (int)((gDataTimeStamp - hstTbl[i].timestamp[p])*1000);
    if (hstTbl[i].isScheduled) delta_t = (int)hstTbl[i].pollingRate * 2;
    if (delta_t < (int)(hstTbl[i].pollingRate-gSystemTick)) continue;
    hstTbl[i].timestamp[n] = gDataTimeStamp;
    hstTbl[i].dataPtr = n;
    if (needToArchiveRecord(i)) WriteRecordToLTS(i);
    hstTbl[i].isScheduled = FALSE;
  }
  if (recordToRemove >= nHistoryRecords - nstdfiles) 
  { /* finished removing old files */
    mday = tm->tm_mday;
    recordToRemove = 0;
  }
  return;
}
int getLastLTSData(int i,int index,void *datbuf,int datfmt,int bufsize)
{
# ifndef NO_LTS
  int hFile = -1,number,cc = 0;
  SINT32 pstart,rlen;
  char fn[16],fileName[128];
  DTYPE d;
  int fmtsize;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
  time_t timestamp,starttime = time(NULL);
  double dtimestamp, datatime = 0;
  struct tm *lt=localtime((time_t *)&starttime);
  BYTE *tmpHistoryDataBuffer;

  if (bufsize < 1) return buffer_too_small;
  if (hstTbl == NULL) return un_allocated;
  if (i < 0 || i >= nHistoryRecords) return not_defined;
  if (hstTbl[i].depthLong < 0) return invalid_data;
  if (hstTbl[i].recordLength > WORK_SIZE) goto err;

  MakeTargetFileName(fn,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  sprintf(fileName,"%s%s",ArchiveLocation,fn);
  rlen = hstTbl[i].recordLength + timestampLength;
  if ((hFile=open(fileName,O_readmode)) == -1) ccerr(file_error);
  datatime = 0;
  pstart = getFilePosition(hFile,hstTbl[i].useStandardSet);
  pstart -= rlen;

  if (lseek(hFile,pstart,SEEK_SET) < 0) ccerr(file_error);
# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(hstTbl[i].recordLength);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  fmtsize = fmtsizeof(hstTbl[i].c.EqmFormatOut);
  d.dFormat = LFMT(hstTbl[i].c.EqmFormatOut);
  d.dArrayLength = 1;
  d.data.bptr = tmpHistoryDataBuffer + index*fmtsize;
  if (useHighResolutionTimeStamp)
  {
    if (read(hFile,(char *)&dtimestamp,timestampLength) != timestampLength) ccerr(file_error);
  }
  else
  {
    if (read(hFile,(char *)&timestamp,timestampLength) != timestampLength) ccerr(file_error);
    dtimestamp = (double)timestamp;
  }
  if (read(hFile,(char *)tmpHistoryDataBuffer,hstTbl[i].recordLength)
      != hstTbl[i].recordLength) ccerr(file_error);
  datatime = dtimestamp;

  switch (datfmt)
  {
    case CF_FLOAT:
      d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
      number = (int)d.dArrayLength;
      cc = GetValuesAsFloat(&d,(float *)datbuf,(short)number);
      break;
    case CF_LONG:
      d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
      number = (int)d.dArrayLength;
      cc = GetValuesAsLong(&d,(SINT32 *)datbuf,(short)number);
      break;
    case CF_DOUBLE:
      d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
      number = (int)d.dArrayLength;
      cc = GetValuesAsDouble(&d,(double *)datbuf,(short)number);
      break;
    default: goto err;
  }
err:
  if (hFile != -1) close(hFile);
  if (datatime > 0) setDataTimeStamp(datatime);
  return cc;
# else
  return 0;
# endif /* NO_LTS */
}
int loadLTSData(int i,int index,double dstarttime,double dendtime,
                 float lwrLimit, float uprLimit,
                 void *datbuf,int datfmt,int bufsize,int sample)
{
# ifndef NO_LTS
  int hFile = -1,nrecs,raster = 1,isSnapshot = 0;
  time_t starttime = (time_t)dstarttime;
  SINT32 pstart,pstop,cnt = 0,rlen;
  char fn[16],fileName[128];
  DTYPE d;
  int number = 0,fmtsize, cc = 0;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
  time_t timestamp;
  double dtimestamp, datatime = 0, dtnow = makeDataTimeStamp();
  static double timebin = -1;
  struct tm *lt=localtime((time_t *)&starttime);
  BYTE *tmpHistoryDataBuffer, tsBuffer[8];
  UINT32 strsiz;

  if (timebin < 0) timebin = useHighResolutionTimeStamp ? 0.01 : 1.0;
  if (bufsize < 1) return 0;
  if (hstTbl == NULL) return FALSE;
  if (i < 0 || i >= nHistoryRecords) return FALSE;
  if (hstTbl[i].depthLong < 0) return FALSE;
  if ((dtnow - dstarttime) < hstTbl[i].depthShort) return FALSE;
  if (hstTbl[i].recordLength > WORK_SIZE) goto err;
  if (index < 0 || index >= (int)hstTbl[i].c.EqmSizeOut) goto err;
  if (((dendtime-dstarttime) < 0.01) && datfmt <= CF_FLOAT && bufsize > 1) isSnapshot = TRUE;
  if (isSnapshot) 
  {
    index = 0; /* snapshot requests always have index = 0 */
    datatime = dstarttime - timebin;
  }

  MakeTargetFileName(fn,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  sprintf(fileName,"%s%s",ArchiveLocation,fn);

  if (dendtime < dstarttime) dendtime = dtnow;
  if ((pstart=locateRecordFromLTS(fileName,i,dstarttime)) < 0) goto err;
  if ((pstop=locateRecordFromLTS(fileName,i,dendtime)) < 0) goto err;
  rlen = hstTbl[i].recordLength + timestampLength;
  nrecs = (int)((pstop-pstart)/rlen);
  if (bufsize == 1) 
  {
    setDataTimeStamp(dendtime);
    return MAX(nrecs,0);
  }
  if (bufsize < nrecs) raster += nrecs/(bufsize+1);
  if (sample > 0) raster = sample;

  if ((hFile=open(fileName,O_readmode)) == -1) goto err;
  number = 0; cnt = 0; 
  lseek(hFile,pstart,SEEK_SET);

# ifdef INTEL
  tmpHistoryDataBuffer = (BYTE *)alloca(hstTbl[i].recordLength);
# else
  tmpHistoryDataBuffer = tmpWorkArea;
# endif
  fmtsize = fmtsizeof(hstTbl[i].c.EqmFormatOut);

  d.dFormat = LFMT(hstTbl[i].c.EqmFormatOut);
  d.dArrayLength = 1;
  d.data.bptr = tmpHistoryDataBuffer + index*fmtsize;

  while(read(hFile,(char *)tsBuffer,timestampLength) == timestampLength)   /* read through file */
  {
    if (useHighResolutionTimeStamp) 
    {
      memcpy(&dtimestamp,tsBuffer,timestampLength); 
      timestamp = (time_t)dtimestamp;
    }
    else 
    {
      memcpy(&timestamp,tsBuffer,timestampLength);
      dtimestamp = (double)timestamp;
    }
    cnt++;
    if (isSnapshot && dtimestamp > dendtime) dendtime = dtimestamp;
    if (dtimestamp > dendtime) goto err;
    if (dtimestamp < datatime + timebin)
    {
      if (lseek(hFile,hstTbl[i].recordLength,SEEK_CUR) < 0) goto err;
      continue;
    }
    if (read(hFile,(char *)tmpHistoryDataBuffer,hstTbl[i].recordLength)
        != hstTbl[i].recordLength) goto err;

    if (ServerExitCondition) goto err;
    datatime = dtimestamp;

    switch (datfmt)
    {
      case CF_FLOAT:
        d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
        number = (int)d.dArrayLength;
        cc = GetValuesAsFloat(&d,(float *)datbuf,(short)number);
        goto err;
      case CF_FLTINT:
        ((FLTINT *)datbuf)[number].ival = (UINT32)timestamp;
        cc = GetValuesAsFloat(&d,&((FLTINT *)datbuf)[number].fval,1);
        if (cc != 0) goto err;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_INTFLTINT: /* CF_TDS */
        ((INTFLTINT *)datbuf)[number].i1val = (UINT32)timestamp;
        cc = GetValuesAsFloat(&d,&((INTFLTINT *)datbuf)[number].fval,1);
        if (cc != 0) goto err;
        if (GetIntValue(&d,&((INTFLTINT *)datbuf)[number].i2val) != 0)
          ((INTFLTINT *)datbuf)[number].i2val = 0;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_FLTINTINT:
        cc = GetValuesAsFloat(&d,&((FLTINTINT *)datbuf)[number].fval,1);
        if (cc != 0) goto err;
        ((FLTINTINT *)datbuf)[number].i1val = (UINT32)timestamp;
        ((FLTINTINT *)datbuf)[number].i2val = 0;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_LONG:
        d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
        number = (int)d.dArrayLength;
        cc = GetValuesAsLong(&d,(SINT32 *)datbuf,(short)number);
        goto err;
      case CF_LNGINT:
        ((INTINT *)datbuf)[number].i2val = (UINT32)timestamp;
        cc = GetValuesAsLong(&d,&((INTINT *)datbuf)[number].i1val,1);
        if (cc != 0) goto err;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_DOUBLE:
        d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
        number = (int)d.dArrayLength;
        cc = GetValuesAsDouble(&d,(double *)datbuf,(short)number);
        goto err;
      case CF_DBLDBL:
        ((DBLDBL *)datbuf)[number].d2val = dtimestamp;
        cc = GetValuesAsDouble(&d,&((DBLDBL *)datbuf)[number].d1val,1);
        if (cc != 0) goto err;
        if (sample < 0 && !isInRange((BYTE *)datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_NAME16I:
        ((NAME16I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 16;
        cc = GetValuesAsString(&d,((NAME16I *)datbuf)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_NAME32I:
        ((NAME32I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 32;
        cc = GetValuesAsString(&d,((NAME32I *)datbuf)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_NAME48I:
        ((NAME48I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 48;
        cc = GetValuesAsString(&d,((NAME48I *)datbuf)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_NAME64I:
        ((NAME64I *)datbuf)[number].ival = (UINT32)timestamp; strsiz = 64;
        cc = GetValuesAsString(&d,((NAME64I *)datbuf)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_SHORT:
        d.dArrayLength = MIN(bufsize,(int)hstTbl[i].c.EqmSizeOut);
        number = (int)d.dArrayLength;
        cc = GetValuesAsShort(&d,(short *)datbuf,(short)number);
        goto err;
      case CF_DBLDBLDBL:
	      ((DBLDBLDBL *)datbuf)[number].d3val = timestamp;
 	      cc = GetValuesAsDBLDBL(&d,&((DBLDBL *)datbuf)[number],1);
        if (cc != 0) goto err;
	      break;
      case CF_NAME64DBLDBL:
	      ((NAME64DBLDBL *)datbuf)[number].d2val = timestamp;
 	      cc = GetValuesAsNAME64DBL(&d,&((NAME64DBL *)datbuf)[number],1);
        if (cc != 0) goto err;
	      break;
      case CF_NAME64DBLDBLDBL:
	      ((NAME64DBLDBLDBL *)datbuf)[number].d3val = timestamp;
 	      cc = GetValuesAsNAME64DBLDBL(&d,&((NAME64DBLDBL *)datbuf)[number],1);
        if (cc != 0) goto err;
	      break;
      default: 
        goto err;
    }
    number++;
    if (number >= bufsize) goto err;
    if (raster > 1) /* move along through file */
    {
      if (lseek(hFile,rlen*(raster-1),SEEK_CUR) < 0) goto err;
    }
  }
err:
  if (hFile != -1) close(hFile);
  if (datatime > 0) setDataTimeStamp(datatime);
  if (cc != 0) return 0;
  return number;
# else
  return 0;
# endif /* NO_LTS */
}
int IsLiteralDevice(char *eqm,char *prp,char *dev)
{
  int i;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE)) continue;
    break;
  }
  return nHistoryRecords > i ? TRUE: FALSE;
}
TINE_EXPORT int setSTSData(char *eqm,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize)
{
  int i,n,fmtsiz,sttptr=0,literalDevice;
  BYTE *dptr=NULL;

  if (nHistoryRecords <= 0 || hstTbl == NULL) return -un_allocated;
  literalDevice = IsLiteralDevice(eqm,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqm,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (hstTbl[i].depthShort == 0) return -un_allocated;
  if (index < 0) return index;
  if (index >= (int)hstTbl[i].c.EqmSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  fmtsiz = fmtsizeof(hstTbl[i].c.EqmFormatOut);
  if (arraysize > hstTbl[i].depthShort) 
  {
    sttptr = arraysize - hstTbl[i].depthShort;
    arraysize = hstTbl[i].depthShort;
  }

  for (n=0; n < arraysize && n <= hstTbl[i].depthShort; n++)
  {
    dptr = hstTbl[i].data + index*fmtsiz + n*hstTbl[i].recordLength;
    switch (fmt)
    {
      case CF_FLTINT: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_FLOAT) return -illegal_format;
        *((float *)dptr) = ((FLTINT *)data)[n+sttptr].fval;
        hstTbl[i].timestamp[n] = (double)((FLTINT *)data)[n+sttptr].ival;
        break;
      case CF_DBLINT:
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_DOUBLE) return -illegal_format;
        *((double *)dptr) = ((DBLINT *)data)[n+sttptr].dval;
        hstTbl[i].timestamp[n] = (double)((DBLINT *)data)[n+sttptr].ival;
        break;
      case CF_LNGINT: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_LONG) return -illegal_format;
        *((UINT32 *)dptr) = ((LNGINT *)data)[n+sttptr].lval;
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_DBLDBL:
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_DOUBLE) return -illegal_format;
        *((double *)dptr) = ((DBLDBL *)data)[n+sttptr].d1val;
        hstTbl[i].timestamp[n] = ((DBLDBL *)data)[n+sttptr].d2val;
        break;
      case CF_NAME16I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME16) return -illegal_format;
        strncpy(((NAME16 *)dptr)->name,((NAME16I *)data)[n+sttptr].name,16);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME32I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME32) return -illegal_format;
        strncpy(((NAME32 *)dptr)->name,((NAME32I *)data)[n+sttptr].name,32);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME48I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME48) return -illegal_format;
        strncpy(((NAME48 *)dptr)->name,((NAME48I *)data)[n+sttptr].name,48);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
      case CF_NAME64I: 
        if (LFMT(hstTbl[i].c.EqmFormatOut) != CF_NAME64) return -illegal_format;
        strncpy(((NAME64 *)dptr)->name,((NAME64I *)data)[n+sttptr].name,64);
        hstTbl[i].timestamp[n] = (double)((LNGINT *)data)[n+sttptr].ival;
        break;
    }
  }    
  hstTbl[i].dataPtr = (arraysize-1) % hstTbl[i].depthShort;
  return n;
}
int getSTSData(char *eqm,char *prp, char *dev,int index,
		 double dstarttime,double dendtime,
     float lwrLimit, float uprLimit,
     BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag)
{
  int cc=0,i,n,number,rnumber=0,num,raster,cnt,fmtsiz,literalDevice;
  time_t timestamp;
  int isSnapshot=0;
  int sttidx = recsreadFlag ? 1 : 0;
  double dtimestamp,lastreadtimestamp=0;
  double depthInSeconds, pps;
  DTYPE d;
  UINT32 strsiz;

  if (nHistoryRecords <= 0 || hstTbl == NULL) return -un_allocated;
  literalDevice = IsLiteralDevice(eqm,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqm,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (hstTbl[i].depthShort == 0) return -getRemoteHistoryServer(eqm,prp);
  if (index < 0) return index;
  if (index >= (int)hstTbl[i].c.EqmSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  pps = 1000.0 / (double)hstTbl[i].pollingRate;
  depthInSeconds = (double)hstTbl[i].depthShort / pps;
  if (dstarttime < dendtime - depthInSeconds) dstarttime = dendtime - depthInSeconds;

  d.dFormat = LFMT(hstTbl[i].c.EqmFormatOut);
  d.dArrayLength = 1;
  if ((raster=(int)((dendtime-dstarttime)*pps/(double)arraysize)) < 1) raster = 1;
  if (arraysize == 1) 
  {
    raster = 1;
    sttidx = 0;
  }
  isSnapshot = ((dendtime-dstarttime) < .01) ? TRUE : FALSE;
  if (isSnapshot && arraysize > 1) index = 0; /* snapshot requests always have index = 0 */
  fmtsiz = fmtsizeof(d.dFormat);
  n = hstTbl[i].dataPtr;
  /* move raster + sttidx past the current position to start at the beginning */
  n = (n + raster + sttidx + hstTbl[i].depthShort) % hstTbl[i].depthShort;
  if (arraysize == (int)(dendtime-dstarttime)) /* caller wants the array filled */
  {
    if (arraysize <= (num=hstTbl[i].depthShort))
    {
      n = (n + (num-arraysize) + hstTbl[i].depthShort) % hstTbl[i].depthShort;
      dstarttime = hstTbl[i].timestamp[n];
      dendtime = makeDataTimeStamp();
    }
  }
  lastreadtimestamp = dstarttime;
  for (number=sttidx,cnt=0; number < arraysize &&
                            number <= hstTbl[i].depthShort &&
                            cnt < hstTbl[i].depthShort; cnt += raster)
  {
    d.data.vptr = hstTbl[i].data + index*fmtsiz + n*hstTbl[i].recordLength;
    dtimestamp = hstTbl[i].timestamp[n];
    n = (n + raster + hstTbl[i].depthShort) % hstTbl[i].depthShort;
    if (isSnapshot && dtimestamp > dendtime) dendtime = dtimestamp;
    if (dtimestamp < lastreadtimestamp || dtimestamp > dendtime) continue;
    lastreadtimestamp = dtimestamp;
    timestamp = (time_t)dtimestamp;
    rnumber++;
    if (arraysize < 2) continue;
    switch (fmt)
    { /* fmt is the requested format; d.dFormat is the stored format */
      case CF_FLOAT:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)hstTbl[i].c.EqmSizeOut);
          rnumber = number = (int)d.dArrayLength;
          cc = GetValuesAsFloat(&d,(float *)data,(short)number);
          goto err;
        }
        cc = GetValuesAsFloat(&d,&((float *)data)[number],1);
        if (cc != 0) goto err;
        break;
      case CF_FLTINT:
        ((FLTINT *)data)[number].ival = (UINT32)timestamp;
        cc = GetValuesAsFloat(&d,&((FLTINT *)data)[number].fval,1);
        if (cc != 0) goto err;
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_INTFLTINT: /* CF_TDS */
        ((INTFLTINT *)data)[number].i1val = (UINT32)timestamp;
        cc = GetValuesAsFloat(&d,&((INTFLTINT *)data)[number].fval,1);
        if (cc != 0) goto err;
        if (GetIntValue(&d,&((INTFLTINT *)data)[number].i2val) != 0)
          ((INTFLTINT *)data)[number].i2val = 0;
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_FLTINTINT: 
        cc = GetValuesAsFloat(&d,&((FLTINTINT *)data)[number].fval,1);
        if (cc != 0) goto err;
        ((FLTINTINT *)data)[number].i1val = (UINT32)timestamp;
        ((FLTINTINT *)data)[number].i2val = (UINT32)((dtimestamp-(double)timestamp)*1000);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_LONG:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)hstTbl[i].c.EqmSizeOut);
          rnumber = number = (int)d.dArrayLength;
          cc = GetValuesAsLong(&d,(SINT32 *)data,(short)number);
          goto err;
        }
        cc = GetValuesAsLong(&d,&((SINT32 *)data)[number],1);
        if (cc != 0) goto err;
        break;
      case CF_LNGINT:
        ((LNGINT *)data)[number].ival = (SINT32)timestamp;
        cc = GetValuesAsLong(&d,&((LNGINT *)data)[number].lval,1);
        if (cc != 0) goto err;
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_DOUBLE:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)hstTbl[i].c.EqmSizeOut);
          rnumber = number = (int)d.dArrayLength;
          cc = GetValuesAsDouble(&d,(double *)data,(short)number);
          goto err;
        }
        cc = GetValuesAsDouble(&d,&((double *)data)[number],1);
        if (cc != 0) goto err;
        break;
      case CF_DBLDBL:
        ((DBLDBL *)data)[number].d2val = dtimestamp;
        cc = GetValuesAsDouble(&d,&((DBLDBL *)data)[number].d1val,1);
        if (cc != 0) goto err;
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_NAME16I:
        ((NAME16I *)data)[number].ival = (SINT32)timestamp; strsiz = 16;
        cc = GetValuesAsString(&d,((NAME16I *)data)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_NAME32I:
        ((NAME32I *)data)[number].ival = (SINT32)timestamp; strsiz = 32;
        cc = GetValuesAsString(&d,((NAME32I *)data)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_NAME48I:
        ((NAME48I *)data)[number].ival = (SINT32)timestamp; strsiz = 48;
        cc = GetValuesAsString(&d,((NAME48I *)data)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_NAME64I:
        ((NAME64I *)data)[number].ival = (SINT32)timestamp; strsiz = 64;
        cc = GetValuesAsString(&d,((NAME64I *)data)[number].name,&strsiz);
        if (cc != 0) goto err;
        break;
      case CF_SHORT:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)hstTbl[i].c.EqmSizeOut);
          rnumber = number = (int)d.dArrayLength;
          cc = GetValuesAsShort(&d,(short *)data,(short)number);
          goto err;
        }        
        return -illegal_format;
      case CF_DBLDBLDBL:
	      ((DBLDBLDBL *)data)[number].d3val = timestamp;
 	      cc = GetValuesAsDBLDBL(&d,&((DBLDBL *)data)[number],1);
        if (cc != 0) goto err;
	      break;
      case CF_NAME64DBLDBL:
	      ((NAME64DBLDBL *)data)[number].d2val = timestamp;
 	      cc = GetValuesAsNAME64DBL(&d,&((NAME64DBL *)data)[number],1);
        if (cc != 0) goto err;
	      break;
      case CF_NAME64DBLDBLDBL:
	      ((NAME64DBLDBLDBL *)data)[number].d3val = timestamp;
 	      cc = GetValuesAsNAME64DBLDBL(&d,&((NAME64DBLDBL *)data)[number],1);
        if (cc != 0) goto err;
	      break;
      default: 
        goto err;
		}
    number++;
  }

  assignNumberOfPoints(data,fmt,arraysize,&number,&rnumber,recsreadFlag);

err:
  if (cc != 0) return -cc;
  if (lastreadtimestamp > 0) setDataTimeStamp((double)lastreadtimestamp);
  return arraysize < 2 ? rnumber : number;
}
int isInRange(BYTE *data,int fmt,int number,float lwrLimit,float uprLimit)
{
  float fval;
  switch (fmt)
  {
    case CF_FLTINT:
      fval = ((FLTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((FLTINT *)data)[number].ival = 0; /* flag the boundary */
        if (number > 0 && ((FLTINT *)data)[number-1].ival == 0) return 0;
      }
      return -1;
    case CF_FLTINTINT:
      fval = ((FLTINTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((FLTINTINT *)data)[number].i1val = 0; /* flag the boundary */
        ((FLTINTINT *)data)[number].i2val = 0; /* flag the boundary */
        if (number > 0 && ((FLTINTINT *)data)[number-1].i1val == 0) return 0;
      }
      return -1;
    case CF_INTINT:
      fval = (float)((INTINT *)data)[number].i1val;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((INTINT *)data)[number].i2val = 0; /* flag the boundary */
        if (number > 0 && ((INTINT *)data)[number-1].i2val == 0) return 0;
      }
      return -1;
    case CF_DBLDBL:
      fval = (float)((DBLDBL *)data)[number].d1val;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((DBLDBL *)data)[number].d2val = 0; /* flag the boundary */
        if (number > 0 && ((DBLDBL *)data)[number-1].d2val == 0) return 0;
      }
      return -1;
    case CF_INTFLTINT:
      fval = (float)((INTFLTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((INTFLTINT *)data)[number].i1val = 0; /* flag the boundary */
        if (number > 0 && ((INTFLTINT *)data)[number-1].i1val == 0) return 0;
      }
      return -1;
  }
  return -1;
}
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound,int recsreadFlag)
{
  int padding = recsreadFlag ? 1 : 0;
  if (data == NULL || npoints == NULL || nfound == NULL) return;

  if (siz == 1) *npoints = *nfound + padding;
  else if (!recsreadFlag) return;

  switch (fmt)
  {
    case CF_FLTINT:
      if (padding) (*npoints)--;
      ((FLTINT *)data)[0].ival = (UINT32)(*npoints);
      ((FLTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_FLTINTINT:
      if (padding) (*npoints)--;
      ((FLTINTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].i2val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_INTINT:
      if (padding) (*npoints)--;
      ((INTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((INTINT *)data)[0].i2val = (UINT32)(*npoints);
      break;
    case CF_DBLDBL:
      if (padding) (*npoints)--;
      ((DBLDBL *)data)[0].d1val = (double)(*npoints);
      ((DBLDBL *)data)[0].d2val = (double)(*npoints);
      break;
    case CF_NAME16I:
      if (padding) (*npoints)--;
      ((NAME16I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME32I:
      if (padding) (*npoints)--;
      ((NAME32I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME48I:
      if (padding) (*npoints)--;
      ((NAME48I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME64I:
      if (padding) (*npoints)--;
      ((NAME64I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_LONG:
      if (siz == 1) *(UINT32 *)data = (UINT32)(*nfound);
      break;
    case CF_FLOAT:
      if (siz == 1) *(float *)data = (float)(*nfound);
      break;
    case CF_INTFLTINT:
      if (siz == 1) 
      {
        ((INTFLTINT *)data)[0].i1val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].i2val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].fval = (float)(*nfound);
      }
      break;
    case CF_DOUBLE:
      if (siz == 1) *(double *)data = (double)(*nfound);
      break;
  }
}

int nDaysInTimeSpan(struct tm *s,struct tm *e)
{
  int n = 0, nx, y, d, ms, me, i;

  d = nDaysInMonth(e->tm_year,e->tm_mon);
  nx = s->tm_mday + (d-e->tm_mday-1);
  for (y=s->tm_year; y<e->tm_year+1; y++)
  {
    ms = y == s->tm_year ? s->tm_mon : 0;
    me = y == e->tm_year ? e->tm_mon : 11;
    for (i=ms; i<me+1; i++) 
    {
      n += nDaysInMonth(y,i);
    }
  }
  return n - nx;
}
time_t getStartTimeFromDepth(char *eqm,char *prp, int depth)
{
  time_t t=time(NULL);
  int recsize,i,timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(SINT32);
# ifndef NO_LTS
  int n,k,siz;
  struct tm *lt,it,itprev;
  char fn[16],fileName[128];
  FILE *fp = NULL;
# endif

  if (nHistoryRecords <= 0 || hstTbl == NULL) return -un_allocated;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  recsize = hstTbl[i].recordLength + timestampLength;
# ifndef NO_LTS
  /* if no long term storage then there's only what's in short term */
  if (hstTbl[i].depthLong == 0) return t - hstTbl[i].depthShort;
  /* start looking backwards from today to find a start point where at least 'depth' points are stored */
  lt = localtime(&t); 
  it = *lt; it.tm_sec = 0; it.tm_hour = 0; it.tm_min = 0; itprev = it;
  for (k=0,n=0; k<30 && n<depth; k++)
  {
    MakeTargetFileName(fn,it.tm_year,it.tm_mon+1,it.tm_mday,i);
    sprintf(fileName,"%s%s",ArchiveLocation,fn);
    if ((fp=fopen(fileName,"r")) != NULL)
    {
      fseek(fp,0,SEEK_END);
      siz = ftell(fp);
      fclose(fp);
      if (siz > 0) 
      {
        if (n + siz/recsize > depth) 
        {
          return n == 0 ? mktime(&it) : mktime(&itprev);
        }
        else 
        {
          n += siz/recsize;
        }
      }
    }
    itprev = it;
    if (it.tm_mday > 1) it.tm_mday--;
    else 
    {
      if (it.tm_mon > 0)
      {
        it.tm_mon--;
      }
      else
      {
        it.tm_year--;
        it.tm_mon = 11;
      }
      it.tm_mday = nDaysInMonth(it.tm_year,it.tm_mon) + 1;
    }  
  }
  if (n > 0) return mktime(&itprev);
# endif
  return t - hstTbl[i].depthShort;
}
int getLTSData(char *eqm,char *prp, char *dev,int index,
               double dstarttime,double dendtime,
               float lwrLimit, float uprLimit,
               BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag)
{
# ifndef NO_LTS
  SINT32 ps;
  int recsread=0,ptr=0,recsfound=0,rnumber=0,recsize,i,asize,daystotal,daysinmonth;
  int literalDevice,finalPass=FALSE;
  struct tm st,et,*lt,it;
  time_t t = time(NULL),itime = 0;
  time_t starttime = (time_t)dstarttime;
  time_t endtime = (time_t)dendtime;

  if (nHistoryRecords <= 0 || hstTbl == NULL) return -un_allocated;
  literalDevice = IsLiteralDevice(eqm,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqm,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (hstTbl[i].depthShort == 0) return -getRemoteHistoryServer(eqm,prp);
  if (hstTbl[i].depthLong == 0) return 0;
  if (starttime >= t - hstTbl[i].depthShort * STS_DEADBAND) return 0;

  if (index < 0) return index;
  if (index > (int)hstTbl[i].c.EqmSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  recsize = fmtsizeof(fmt);
  lt = localtime(&endtime); et = *lt;
  lt = localtime(&starttime); st = *lt;
  if (et.tm_mon == st.tm_mon && et.tm_mday == st.tm_mday)
  {
    asize = arraysize;
    finalPass = TRUE;
  }
  else if (arraysize == 1)   /* just a number query */
  {
    asize = arraysize;
  }
  else                      /* spans multiple files */
  {
    asize = 0;
  }

  daystotal = nDaysInTimeSpan(&st,&et);

  if (recsreadFlag && arraysize > 1)
  {
    recsread = 1;
    ptr = recsread * recsize;
  }
  while (recsread < arraysize)
  {
    if (asize < arraysize || arraysize == 1)
    {
      it = st;
      if (it.tm_year == et.tm_year && it.tm_mon == et.tm_mon && 
          (it.tm_mday == et.tm_mday || useMonthlyHistoryFiles))
      {
        asize = arraysize;
        finalPass = TRUE;
      }
      else  /* move along */
      {
        it.tm_mday = (it.tm_mday+1)%(nDaysInMonth(it.tm_year,it.tm_mon)+1);
        if (it.tm_mday == 0)
        {
          it.tm_mday++;
          it.tm_mon = (it.tm_mon+1)%12;
          if (it.tm_mon == 0) it.tm_year++;
        }
        it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
        itime = mktime(&it);
        daysinmonth = (int)MAX((itime-starttime)/(3600*24),1);
        if (arraysize > 1) 
        {
          if (fmt <= CF_FLOAT) 
          {
            asize = arraysize;
          }
          else 
          {
            asize += MAX((arraysize * daysinmonth)/daystotal,1);
          }
          if (asize >= arraysize) finalPass = TRUE; /* just in case */
        }
      }
    }
    if (asize > arraysize) asize = arraysize;
    ps = loadLTSData(i,index,dstarttime,dendtime,lwrLimit,uprLimit,&data[ptr],fmt,asize-recsread,sample);
    if (asize == 1) recsfound += (int)ps; else recsread += (int)ps;
    ptr = recsread * recsize;
    if (finalPass) break;
    starttime = itime;
    dstarttime = (double)starttime;
    lt = localtime(&starttime); st = *lt;
  }
  rnumber = recsfound > 0 ? recsfound : recsread;
  if (arraysize == 1) recsread = recsfound;

  assignNumberOfPoints(data,fmt,arraysize,&recsread,&recsfound,recsreadFlag);

  return rnumber;
# else
  return 0;
# endif /* NO_LTS */
}
int getRemoteHistoryServer(char *eqm,char *prp)
{
  int i;
  ExportListStruct *el = getExportListItem(eqm);
  if (nHistoryRecords <= 0 || hstTbl == NULL) return un_allocated;
  if (el == NULL) return non_existent_elem;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) return un_allocated;
  if (hstTbl[i].prd == NULL) return not_defined;
  memcpy(el->EqmLastErrorString,hstTbl[i].prd->prpRedirection,STATUS_SIZE);
  return server_redirection;
}
#ifndef SMALL_TINE_LIB
int findAvailableHistoryIndex(void)
{
  int i,maxidx=0;

  for (i=0; i<nHistoryRecords; i++)
  {
    if (hstTbl[i].recordIndex > maxidx) maxidx = hstTbl[i].recordIndex;
  }
  return maxidx + 1;
}
#endif /* !SMALL_TINE_LIB */
#endif /* NO_DOXYGEN */
#ifndef SMALL_TINE_LIB
TINE_EXPORT int AppendHistoryInformation(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec)
{
  int i, tblidx, cc = 0;
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);
  static int isLateAddition = FALSE;

  if (eqm == NULL || prp == NULL || dev == NULL || hspec == NULL) ccerr(argument_list_error);

  historyInit(GetExportedName(eqm),eqm);

  if (hstTbl == NULL)
  {
    if ((hstTbl=(HstTblEntry *)SystemCalloc(maxHistoryRecords,sizeof(HstTblEntry))) == NULL)
      ccerr(out_of_local_memory);
  }

  if ((i=nHistoryRecords) >= maxHistoryRecords) ccerr(resources_exhausted);
  if (!isLateAddition && (t - startuptime > 300))
  { /* 5 minutes after server start */
    isLateAddition = TRUE;
  }

  hstTbl[i].recordIndex = idx;
  strncpy(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE);
  strncpy(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE);
  hstTbl[i].c.EqmSizeOut = len;
  hstTbl[i].c.EqmFormatOut = BFMT(fmt);
  if ((tblidx=getHistoryTableIndex(&hstTbl[i],nHistoryRecords)) >= 0)
  { /* already in the history tables ! */
    i = tblidx;
    hstTbl[i].heartbeat = hspec->heartbeat;
    hstTbl[i].pollingRate = hspec->pollingRate;
    if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate *= 1000; /* was thinking in seconds ! */
    if (hstTbl[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = hstTbl[i].pollingRate;
    hstTbl[i].pTolerance = hspec->pTolerance;
    if (hstTbl[i].pTolerance > 1.0) hstTbl[i].pTolerance /= 100;
    hstTbl[i].aTolerance = hspec->aTolerance;
    hstTbl[i].depthLong = hspec->depthLong;
    feclog("HST: modifying history parameters for entry %s %s %s (%d)",
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmProperty,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].recordIndex);
    ccerr(0);
  }
  hstTbl[i].heartbeat = hspec->heartbeat;
  hstTbl[i].pollingRate = hspec->pollingRate;
  if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate *= 1000; /* was thinking in seconds ! */
  if (hstTbl[i].pollingRate < 20) hstTbl[i].pollingRate = 1000;  /* don't allow some idiot to pass a zero ! */
  if (hstTbl[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = hstTbl[i].pollingRate;
  hstTbl[i].archiveRate = hspec->archiveRate;
  if (hstTbl[i].archiveRate < 20) hstTbl[i].archiveRate *= 1000; /* was thinking in seconds ! */
  if (hstTbl[i].archiveRate < 20) hstTbl[i].archiveRate = 1000; /* don't allow some idiot to pass a zero ! */
  hstTbl[i].pTolerance = hspec->pTolerance;
  if (hstTbl[i].pTolerance > 1.0) hstTbl[i].pTolerance /= 100;
  hstTbl[i].aTolerance = hspec->aTolerance;
  hstTbl[i].depthShort = hspec->depthShort;
  hstTbl[i].depthLong = hspec->depthLong;
  hstTbl[i].recordLength =
      (int)hstTbl[i].c.EqmSizeOut * fmtsizeof(hstTbl[i].c.EqmFormatOut);
# ifdef OS_16BIT
  if (hstTbl[i].recordLength > 0x3fff)
  {
    feclog("HIST: %s %s %s record too long (%d bytes)",
            hstTbl[i].c.EqmName,
            hstTbl[i].c.EqmProperty,
            hstTbl[i].c.EqmDeviceName,
            hstTbl[i].recordLength);
    ccerr(resources_exhausted);
  }
# endif
  MakeTargetFileName(hstTbl[i].ArchiveFileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  hstTbl[i].useStandardSet = gIsUsingStdFileSet && !isLateAddition ? TRUE : FALSE;
  if (hspec->rhsServerName != NULL && strlen(hspec->rhsServerName) > 0)
  {
    if ((hstTbl[i].prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) 
      ccerr(out_of_server_memory);

    strncpy(&hstTbl[i].prd->prpRedirection[PRP_REDIR_SRV_OFFSET],hspec->rhsServerName,16);
    if (hspec->rhsPropertyName != NULL) 
      strncpy(&hstTbl[i].prd->prpRedirection[PRP_REDIR_PRP_OFFSET],hspec->rhsPropertyName,32);
  }
  if ((hstTbl[i].timestamp=(double *)SystemCalloc(hstTbl[i].depthShort,sizeof(double))) == NULL)
      ccerr(out_of_local_memory);
  if ((hstTbl[i].data=(BYTE *)SystemCalloc(hstTbl[i].depthShort,hstTbl[i].recordLength)) == NULL)
      ccerr(out_of_local_memory);
  feclog("HIST: %s %s %s (%d s %d mon)",
          hstTbl[i].c.EqmName,
          hstTbl[i].c.EqmProperty,
          hstTbl[i].c.EqmDeviceName,
          hstTbl[i].depthShort,
          hstTbl[i].depthLong);

  nHistoryRecords++;
  if (hstTbl[i].useStandardSet) makeFilesForRecord(i);
  if (MaxPollingRate > minHistoryPollingRate)
  {
    SysPoll = minHistoryPollingRate;
    MaxPollingRate = minHistoryPollingRate;
    if (minHistoryPollingRate < MinPollingRate) MinPollingRate = minHistoryPollingRate;
  }
err:
  return cc;
}
TINE_EXPORT int GetLastStoredData(char *eqm,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize)
{
  int ptr=0,i,literalDevice;

  if (nHistoryRecords <= 0 || hstTbl == NULL) return un_allocated;
  literalDevice = IsLiteralDevice(eqm,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(hstTbl[i].c.EqmName,eqm,EQM_NAME_SHORTSIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(hstTbl[i].c.EqmDeviceName,dev,DEVICE_NAME_SIZE))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqm,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return un_allocated;
  if (hstTbl[i].depthLong == 0) return 0;
  if (index < 0) return -index;
  if (index > (int)hstTbl[i].c.EqmSizeOut) return out_of_range;
  if (arraysize < 1) return buffer_too_small;

  return getLastLTSData(i,index,&data[ptr],fmt,arraysize);
}
#endif /* SMALL_TINE_LIB */
