/**
 * \internal
 * \file swaplib.c
 * \brief Byte-Swapping and Structures
 * 
 * This module contains
 * <ul>
 * <li>functions for byte-swapping</li>
 * <li>functions which handle structure descriptions and tagged structures</li>
 * <li>an initializer for globally used structure definitions</li>
 * </ul>
 * 
 * 	 swaplib.c is also to be used when machines either do not use
 * 	 INTEL byte order or use INTEL byte order but have non-BYTE
 * 	 alignment (such as OSF)
*/
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "globals.h"
#include "swaplib.h"
#include "tinetype.h"
#include "errors.h"
#include "hstlib.h"

/*----------------------------------------------------------------------------*
 *  Switches between Intel and Motorola byte order conventions 15.09.93 swh 
 *  Special KLUGE for VMS - deliver Mot order for float,double 10.03.94 swh
 *  Let VMS compile using IEEE format pass thru SwapBytes      10.09.2006 pdd
 *----------------------------------------------------------------------------*/

#if !defined(INTEL) && !defined(NONALIGNED) && !defined(VMS_IEEE)
int SwapBytes(char *dat,int size,int fmt,int dir)
{
  register int i,tmp;
# ifdef VMS
  int j, a = (dir==RTOL?0:1),b = (a+1)%2;
  char c;
# endif
  hdrSPECTRUM shs;

  switch(fmt)
  {
    case CF_NULL: 
      return 0;
    case CF_TEXT:
    case CF_BYTE:
    case CF_BITFIELD8:
    case CF_XML:
    case CF_STRING8:
    case CF_STRING16:
    case CF_STRING32:
    case CF_STRING48:
    case CF_STRUCT:
      return 0;
    case CF_BITFIELD16:
    case CF_SHORT:
#     ifndef VMS
      for (i=0; i<size; i++) _ISWAP(((short *)dat)[i]);
#     endif
	    return 0;
    case CF_BIT:
    case CF_BOOLEAN:
    case CF_BITFIELD32:
    case CF_LONG:
#     ifdef VMS
      return 0;
#     endif
    case CF_FLOAT:
#     ifndef VMS
      for (i=0; i<size*4; i+=4) _LSWAP(dat[i]);
#     else
      for (i=0; i<size; i++) if (((int *)dat)[i]) for (j=0; j<2; j++)
      {
        c = dat[i*4+j+2];
        dat[i*4+j+2] = dat[i*4+j] - a*j;
        dat[i*4+j] = c + b*j;
      }
#     endif
      return 0;
    case CF_INT64:
    case CF_DOUBLE:
#     ifndef VMS
      for (i=0; i<size*8; i+=8) _DSWAP(dat[i]);
#     else
      for (i=0; i<size; i+=8) for (j=0; j<4; j++)
      {
        c= *(dat+i+8+j-2-(j/2));
        *(dat+i+8+j-2-(j/2))= *(dat+i+j);         /*not for VMS serv yet */
        *(dat+i+j)=c + b*(j ? 0 : 0x20);
      }
#     endif
      return 0;
#   ifndef VMS
    case CF_DBLDBL:
      for (i=0; i<size*16; i+=16) {_DSWAP(dat[i]); _DSWAP(dat[i+8]);}
      return 0;
    case CF_NAMEDBLDBL:
      for (i=0; i<size*32; i+=32) {_DSWAP(dat[16+i]); _DSWAP(dat[16+i+8]);}
      return 0;
    case CF_NAME32DBLDBL:
      for (i=0; i<size*48; i+=48) {_DSWAP(dat[32+i]); _DSWAP(dat[32+i+8]);}
      return 0;
    case CF_NAME64DBL:
      for (i=0; i<size*72; i+=72) {_DSWAP(dat[64+i]);}
      return 0;
    case CF_NAME64DBLDBL:
      for (i=0; i<size*80; i+=80) {_DSWAP(dat[64+i]); _DSWAP(dat[64+i+8]);}
      return 0;
    case CF_NAME64DBLDBLDBL:
      for (i=0; i<size*88; i+=88) {_DSWAP(dat[64+i]); _DSWAP(dat[64+i+8]); _DSWAP(dat[64+i+16]);}
      return 0;
    case CF_NAME16FI:
    case CF_NAME16II:
      for (i=0; i<size*24; i+=24) {_LSWAP(dat[i+16]); _LSWAP(dat[i+20]);}
      return 0;
    case CF_NAME8I:
      for (i=0; i<size*12; i+=12) {_LSWAP(dat[i+8]);}
      return 0;
    case CF_NAME16I:
      for (i=0; i<size*20; i+=20) {_LSWAP(dat[i+16]);}
      return 0;
    case CF_NAME32I:
      for (i=0; i<size*36; i+=36) {_LSWAP(dat[i+32]);}
      return 0;
    case CF_NAME48I:
      for (i=0; i<size*52; i+=52) {_LSWAP(dat[i+48]);}
      return 0;
    case CF_NAME64I:
      for (i=0; i<size*68; i+=68) {_LSWAP(dat[i+64]);}
      return 0;
    case CF_FF:
    case CF_FI:
    case CF_II:
      for (i=0; i<size*8; i+=8) {_LSWAP(dat[i]); _LSWAP(dat[i+4]);}
      return 0;
    case CF_III:
    case CF_FII:
    case CF_FFI:
    case CF_TDS:
      for (i=0; i<size*12; i+=12) {_LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]);}
      return 0;
    case CF_DBLDBLDBL:
      for (i=0; i<size*24; i+=24) {_DSWAP(dat[i]); _DSWAP(dat[i+8]); _DSWAP(dat[i+16]);}
      return 0;
    case CF_USTRING:
      for (i=0; i<size*96; i+=96)
      {
        _LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]); _LSWAP(dat[i+12]);
      }
      return 0;
    case CF_UNAME:
      for (i=0; i<size*32; i+=32)
      {
        _LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]); _LSWAP(dat[i+12]);
      }
      return 0;
    case CF_TTII:
    case CF_IIII:
    case CF_IIFF:
    case CF_FIFI:
    case CF_IFFF:
      for (i=0; i<size*16; i+=16)
      {
        _LSWAP(dat[i]); _LSWAP(dat[i+4]); _LSWAP(dat[i+8]); _LSWAP(dat[i+12]);
      }
      return 0;
    case CF_SPECTRUM:
      memcpy(&shs,dat,sizeof(hdrSPECTRUM));
      _StructSwap(SpectrumHeaderStruct,(unsigned char *)dat,(unsigned char *)&shs,dir);
      for (i=96; i<size*4; i+=4) _LSWAP(dat[i]);
      return 0;
#   endif
    default: if (tineDebug > 1) dbglog("SwapBytes : bad format size");
  }
  return illegal_format;
}
#endif  /* !INTEL && !NONALIGNED && !VMS_IEEE */

static int gControlStructInitialized = FALSE;
struct structStruct *structList = NULL;

struct structStruct *getStructFromTag(char *tag)
{
  struct structStruct *s;

  if (tag == NULL || strlen(tag) == 0) return NULL;
  for (s=structList; s && strncmp(s->name,tag,TAG_NAME_SIZE); s=s->next);
  if (s != NULL && (s->rawlen == 0 || s->size == 0))
  {
    if (tineDebug) dbglog("getStructFromTag : struct %s not properly terminated",tag);
    return NULL;
  }
  return s;
}

/**
 * \internal
 * \brief Finds a tagged structure.
 */
struct structStruct *findstruct(char *tag)
{
  struct structStruct *s;
  if (tag == NULL) return NULL;
  for (s=structList; s; s=s->next)
    if (!strncmp(s->name,tag,TAG_NAME_SIZE)) return s;
  return s;
}
/**
 * \internal
 * \brief Finds or allocates a tagged structure.
 */
struct structStruct *getTaggedStruct(char *tag)
{
  struct structStruct *s;

  if ((s=findstruct(tag)) != NULL) return s; /* found it */
  if ((s=(struct structStruct *)SystemCalloc(1,sizeof(struct structStruct)))
       != NULL)
  { /* didn't find it -> need to allocated it */
    strncpy(s->name,tag,TAG_NAME_SIZE);
    s->rawlen = 0;
    s->next = structList;
    structList = s;
    if (tineDebug) dbglog("getTaggedStruct : New structure tag registered: %s",tag);
  }
  return s;
}
int queryStructFormat(char *tag,BYTE *sfq,int fmt,int *sfqsiz)
{
  int i,n,fsz,sz=0,nq,cc=0;
  char field[64];
  struct structStruct *s, *ss;
  struct structFormat *sf;
  NAME64DBLDBL *ndd = (NAME64DBLDBL *)sfq;
  NAME16II *nii = (NAME16II *)sfq;
  INTINT *ii = (INTINT *)sfq;
  
  if (sfqsiz == NULL) ccerr(argument_list_error);
  nq = *sfqsiz;
  if ((s=getStructFromTag(tag)) == NULL) ccerr(invalid_structure_tag);
  for (sf=s->s,n=0; sf != NULL; sf=sf->nxt,n++); /* count them first (reverse ordered!) */
  for (i=0,sf=s->s; i<nq && i<n && sf != NULL; i++, sf=sf->nxt)
  {
    if (sf->fmt == CF_STRUCT)
    {
      if ((ss=getStructFromTag(sf->stag)) == NULL) return invalid_structure_tag;
      fsz = ss->rawlen;
      sprintf(field,"<%.16s>%.16s",sf->stag,sf->field);
    }
    else
    {
      fsz = fmtsizeof(sf->fmt);
      strncpy(field,sf->field,16);
    }
    sz += sf->siz * fsz;
    switch (fmt)
    {
      case CF_INTINT:
        ii[n-i-1].i1val = sf->siz;
        ii[n-i-1].i2val = sf->fmt;
        break;
        break;
      case CF_NAME16II:
        strncpy(nii[n-i-1].name,field,16);
        nii[n-i-1].i1val = sf->siz;
        nii[n-i-1].i2val = sf->fmt;
        break;
      case CF_NAME64DBLDBL:
        strncpy(ndd[n-i-1].name,field,64);
        ndd[n-i-1].d1val = (double)sf->siz;
        ndd[n-i-1].d2val = (double)sf->fmt;
        break;
      default: return illegal_format;
    }
  }
  if (i == nq) i--;
  switch (fmt)
  {
    case CF_INTINT:
      ii[i].i1val = sz;
      ii[i].i2val = CF_NULL;
      break;
      break;
    case CF_NAME16II:
      strncpy(nii[n-i-1].name,tag,16);
      nii[i].i1val = sz;
      nii[i].i2val = CF_NULL;
      break;
    case CF_NAME64DBLDBL:
      strncpy(ndd[i].name,tag,16);
      ndd[i].d1val = (double)sz;
      ndd[i].d2val = (double)CF_NULL;
      break;
    default: return illegal_format;
  }
  *sfqsiz = n+1;
err:
  return cc;
}

/**
 * \brief Prepend a structFormat to a structFomat list.
 * 
 * This method creates a new structFormat and sets it in
 * the first element of a linked list of structFormats.
 * A structFormat may describe a scalar or an array.
 * 
 * 
 * \param target void pointer which is interpreted as a double pointer to a structFormat. Insert the new struct format before this.
 * \param addr Address offset of this field in the structure. 
 * \param size Array length if the described field. (not the size in bytes).
 * \param fmt TINE CF_ format of the new struct element
 * \param subtag is a possible structure tag for the case fmt == CF_STRUCT
 * \param field is the corresponding registered field name
 * 
 * \todo target should be structFormat double pointer.
 */ 
int addtostruct(void *target,int addr,int size, int fmt, char *subtag, char *field)
{
  struct structStruct *ss;
  struct structFormat *tmp,**s = (struct structFormat **)target;
  int offset = (*s) ? (*s)->off + (*s)->siz*getFormatSize((*s)->fmt) : 0;

  if ((*s) != NULL && (*s)->fmt == CF_STRUCT)
  { /* embedded struct -> find the proper offset */
    if ((ss=findstruct((*s)->stag)) == NULL) return illegal_format;
    offset = (*s)->off + (*s)->siz*ss->rawlen;
  }
  tmp = (struct structFormat *)SystemCalloc(1,sizeof(struct structFormat));
  if (tmp == NULL) return out_of_local_memory;
  tmp->off = offset;
  tmp->fmt = fmt;
  tmp->siz = size;
  tmp->addr= addr;
  strncpy(tmp->field,field,16);
  if (subtag != NULL) strncpy(tmp->stag,subtag,16);
  
  /* Add tmp to list */
  tmp->nxt = *s;  
  *s = tmp;
  return 0;
}

/**
 * \internal
 * \brief Frees a structFormat list
 * \param sf Pointer to the first list element
 */
void freeStruct(struct structFormat *sf)
{
  struct structFormat *tmp,*tmp2;
  for (tmp=sf; tmp != NULL; tmp=tmp2)
  {
    tmp2 = tmp->nxt;
    SystemFree(tmp); 
  } 
}
void freeStructList(void)
{
  struct structStruct *s=structList,*snxt;
  while (s != NULL)
  {
#   ifndef DOS
    if (s->sdata != NULL) SystemFree(s->sdata);
#   endif
    freeStruct(s->s);
    snxt = s->next;
    SystemFree(s);
    s = snxt;
  }
  structList = NULL;
}
/**
 * \deprecated
 * \brief Adds a field description to a tagged structure.
 * 
 * This function has two meanings:
 * <ol>
 * <li>Adds a field to a tagged structure</li>
 * <li>Sets the size and the array size of a tagged structure</li>
 * </ol>
 * 
 * \param tag Tag name
 * \param addr <ol><li>Address offset of this field in the structure.</li>
 *  <li>Size of the element in bytes</li></ol>
 * \param size <ol><li>Array length if the described field. (not the size in bytes).</li>
 *  <li>Number of elements</li></ol>
 * \param fmt TINE CF_ format of the new struct element
 * 
 * 
 * \return 0 or out_of_local_memory if no struct could be found/allocated for tag.
 * 
 * \see StructStruct
 * \see findstruct
 * \todo Better use two separate functions than two meanings in one function.
 */ 
TINE_EXPORT int fillstruct(char *tag,int addr,int size,int fmt)
{
  struct structStruct *s;

  if (!gControlStructInitialized)
  { /* if this legacy call is used, make sure the control structs are registered properly */
    initControlStructs();
  }
  if ((s=getTaggedStruct(tag)) == NULL) return out_of_local_memory;
  if (s->sealed) return struct_sealed;
  if (fmt == CF_NULL)
  {
    s->num = size;
    s->size = addr;
#   ifndef DOS
    if ((s->sdata=(BYTE *)SystemCalloc(size,s->size)) == NULL) return out_of_local_memory;
#   ifdef INTEL
    if (s->rawlen == 0) s->rawlen = s->size; /* a quick and dirty register from an intel machine */
#   endif
#   else
    s->rawlen = s->size;
#   endif
    s->sealed = TRUE;
    return 0;
  }
  if (!s->sealed) s->rawlen += size*getFormatSize(fmt);
  return addtostruct(&s->s,addr,size,fmt,NULL,"");
}

/**
 * \brief Adds a field description to a tagged structure.
 * 
 * This function has two meanings:
 * <ol>
 * <li>Adds a field to a tagged structure</li>
 * <li>Sets the size and the array size of a tagged structure</li>
 * </ol>
 * 
 * \param tag structure tag name
 * \param addr <ol><li>address offset of this field in the structure.</li>
 *  <li>Size of the element in bytes</li></ol>
 * \param size <ol><li>array length if the described field. (not the size in bytes).</li>
 *  <li>Number of elements</li></ol>
 * \param fmt format of the new struct element
 * \param field field name
 * 
 * \return 0 or out_of_local_memory if no struct could be found/allocated for tag.
 * 
 * \see StructStruct
 * \see findstruct
 */ 
TINE_EXPORT int AddFieldToStruct(char *tag,int addr,int size,int fmt,char *field)
{
  struct structStruct *s, *ss;
  char *c, *fld, subtag[64];

  if ((s=getTaggedStruct(tag)) == NULL) return out_of_local_memory;
  if (s->sealed) return struct_sealed;
  fld = field;
  if (fmt == CF_STRUCT)
  { /* nested structure */
    if ((c=strchr(field,'<')) != NULL) 
    {
      strncpy(subtag,&field[1],64);
      if ((c=strchr(subtag,'>')) == NULL) return invalid_field;
      *c++ = 0; fld = c;
    }
    else
    {
      strncpy(subtag,field,64);
    }
    if ((ss=findstruct(subtag)) == NULL) return illegal_format;
    s->rawlen += size*ss->rawlen;
  }
  else
  {
    subtag[0] = 0;
    s->rawlen += size*getFormatSize(fmt);
  }
  return addtostruct(&s->s,addr,size,fmt,subtag,fld);
}
/**
 * \brief Seals a tagged structure (registration finished!).
 * 
 * 
 * \param tag structure tag name
 * \param size total size of the structure in bytes
 * \param number number of elements of this structure to allocate
 * 
 * \return 0, un_allocated, out_of_local_memory 
 * 
 * \see StructStruct
 * \see findstruct
 */ 
TINE_EXPORT int SealTaggedStruct(char *tag,int size,int number)
{
  struct structStruct *s;

  if ((s=findstruct(tag)) == NULL) return un_allocated;
  s->num = number;
  s->size = size;
# ifndef DOS
  if ((s->sdata=(BYTE *)SystemCalloc(number,s->size)) == NULL) return out_of_local_memory;
# else
  s->rawlen = s->size;
# endif
  s->sealed = TRUE;
  return 0;
}

TINE_EXPORT int GetStructSize(char *tag)
{
  struct structStruct *s=findstruct(tag);
  return s == NULL ? 0 : s->rawlen;
}
int _StructSwap(struct structFormat *s,unsigned char *dat,unsigned char *struc,int dir)
{
  int n,k,dsiz;
  struct structStruct *ss;

  if (dir == RTOL)
  { /* in-coming */
    for (n=0; s != NULL && s->siz; n += dsiz)
    {
      if (s->fmt == CF_STRUCT)
      { /* nested struct swap */
        if ((ss=findstruct(s->stag)) == NULL) return invalid_structure_tag;
        for (k=0; k<s->siz; k++)
        {
          _StructSwap(ss->s,&dat[s->off+k*ss->rawlen],&struc[s->addr+k*ss->size],RTOL);
        }
        dsiz = s->siz * ss->rawlen;
      }
      else
      {
        SwapBytes((char *)&dat[s->off],s->siz,s->fmt,RTOL);
        memcpy(&struc[s->addr],&dat[s->off],dsiz=s->siz*getFormatSize(s->fmt));
      }
      s = s->nxt;
    }
  }
  else
  { /* out-going  */
    for (n=0; s != NULL && s->siz; n += dsiz)
    {
      if (s->fmt == CF_STRUCT)
      { /* nested struct swap */
        if ((ss=findstruct(s->stag)) == NULL) return invalid_structure_tag;
        for (k=0; k<s->siz; k++)
        {
          _StructSwap(ss->s,&dat[s->off+k*ss->rawlen],&struc[s->addr+k*ss->size],LTOR);
        }
        dsiz = s->siz * ss->rawlen;
      }
      else
      {
        memcpy(&dat[s->off],&struc[s->addr],dsiz=s->siz*getFormatSize(s->fmt));
        SwapBytes((char *)&dat[s->off],s->siz,s->fmt,LTOR);
      }
      s = s->nxt;
    }
  }
  return n;  /* number of bytes swapped */
}

int StructSwap(char *tag,unsigned char *dat,unsigned char *struc,int dir)
{
  struct structStruct *ss=findstruct(tag);
  if (ss == NULL) return 0;
# ifdef DOS
  if (dir == RTOL) memcpy(struc,dat,ss->rawlen);
  else memcpy(dat,struc,ss->rawlen);
  return ss->rawlen;
# else
  return _StructSwap(ss->s,dat,struc,dir);
# endif
}

int StructSwapEx(char *tag,unsigned char *dat,unsigned char *struc,int dir,int num,int structsiz)
{
  int i,n=0;
  struct structStruct *ss=findstruct(tag);
# ifdef INTEL
  int domemcpy;
# endif

  if (ss == NULL || ss->rawlen == 0) return -un_allocated;
  if (num <1 || structsiz < 1) return -out_of_range;
# ifdef INTEL
  domemcpy = ss->s == NULL ? TRUE : FALSE;
# ifdef MSDOS
  domemcpy = TRUE;
# endif
  if (domemcpy)
  { /* tag registered without fields (not a good thing to do, but it should still work) */
    if (dir == RTOL) memcpy(struc,dat,ss->rawlen*num);
    else memcpy(dat,struc,ss->rawlen*num);
    return ss->rawlen*num;
  }
# endif
  for (i=0; i<num; i++)
    n += _StructSwap(ss->s,&dat[i*ss->rawlen],&struc[i*structsiz],dir);
  return n;
}
int _StructSwapEx(struct structFormat *sf,unsigned char *dat,unsigned char *struc,int dir,int num,int structsiz)
{
  int len = 0, fsiz;
  struct structFormat *s = sf;
  struct structStruct *ss;
  int i,n;
# ifdef INTEL
  int domemcpy = (s == NULL) ? TRUE : FALSE;
# endif

  if (num < 1 || structsiz < 1) return -out_of_range;
  while (s != NULL && s->siz)
  { /* run thru the fields */
    fsiz = getFormatSize(s->fmt);
    if (s->fmt == CF_STRUCT)
    { /* embedded struct */
      if ((ss=findstruct(s->stag)) == NULL) return invalid_structure_tag;
      fsiz = ss->rawlen;
    }
    len += s->siz*fsiz;
    s = s->nxt;
  }
# ifdef INTEL
# ifdef MSDOS
  domemcpy = TRUE;
# endif
  if (domemcpy)
  { /* tag registered without fields (not a good thing to do, but it should still work) */
    if (dir == RTOL) memcpy(struc,dat,len*num);
    else memcpy(dat,struc,len*num);
    return len*num;
  }
# endif
  for (i=0,n=0; i<num; i++)
    n += _StructSwap(sf,&dat[i*len],&struc[i*structsiz],dir);
  return n;
}

/**
 * \internal
 * \brief Structure description for HeaderStruct
 * \see HeaderStruct
 */
struct structFormat *HeaderStructStruct;

/**
 * \internal
 * \brief Structure description for CONTRACT
 * \see CONTRACT
 */
struct structFormat *CONTRACTStruct;
struct structFormat *CONTRACT5Struct;

/**
 * \internal
 * \brief Structure description for ProducerHeaderStruct
 * \see ProducerHeaderStruct
 */
struct structFormat *ProducerHeaderStructStruct;
struct structFormat *ProducerHeaderStruct5Struct;

/**
 * \internal
 * \brief Structure description for SUBSCRIPTION
 * \see SUBSCRIPTION
 */
struct structFormat *SUBSCRIPTIONStruct;

/**
 * \internal
 * \brief Structure description for SUBSCRIPTION5
 * \see SUBSCRIPTION5
 */
struct structFormat *SUBSCRIPTION5Struct;

/**
 * \internal
 * \brief Structure description for RPCDataStruct
 * \see RPCDataStruct
 */
struct structFormat *RPCDataStructStruct;

/**
 * \internal
 * \brief Structure description for RPCFecStruct
 * \see RPCFecStruct
 */
struct structFormat *RPCFecStructStruct;

/**
 * \internal
 * \brief Structure description for AMSStruct
 * \see AMSStruct
 */
struct structFormat *AMSStructStruct;

/**
 * \internal
 * \brief Legacy Structure description for AMSStruct
 * \see AMSStruct
 */
struct structFormat *AMS5StructStruct;

/**
 * \internal
 * \brief Structure description for ADSStruct
 * \see ADSStruct
 */
struct structFormat *ADSStructStruct;

/**
 * \internal
 * \brief Legacy Structure description for ADSStruct
 * \see ADSStruct
 */
struct structFormat *ADS5StructStruct;

/**
 * \internal
 * \brief Structure description for AlarmWatchStruct
 * \see AlarmWatchStruct
 */
struct structFormat *AWSStructStruct;

/**
 * \internal
 * \brief Structure description for PropertyQueryStruct
 * \see PropertyQueryStruct
 */
struct structFormat *PropertyQueryStructStruct;

/**
 * \internal
 * \brief Structure description for PropertyQueryStructEx
 * \see PropertyQueryStructEx
 */
struct structFormat *PropertyQueryStructExStruct;

/**
\internal
\brief Structure description for XPropertyQueryStruct
\see XPropertyQueryStruct
*/
struct structFormat *XPropertyQueryStructStruct;

/**
\internal
\brief Structure description for PrpQueryStruct
\see PrpQueryStruct
*/
struct structFormat *PrpQueryStructStruct;

/**
 * \internal
 * \brief Structure description for ActivityQueryStruct
 * \see ActivityQueryStruct
 */
struct structFormat *ActivityQueryStructStruct;

/**
 * \internal
 * \brief Structure description for ServerStateStruct
 * \see ServerStateStruct
 */
struct structFormat *ServerStateStructStruct;

/**
 * \internal
 * \brief Structure description for GlobalDataHdr
 * \see GlobalDataHdr
 */
struct structFormat *GlobalDataHdrStruct;

/**
 * \internal
 * \brief Structure description for ClientQueryStruct
 * \see ClientQueryStruct
*/
struct structFormat *ClientQueryStructStruct;

/**
 * \internal
 * \brief Structure description for SpectrumHeader
 * \see SpectrumHeader
 */
struct structFormat *SpectrumHeaderStruct;

/**
 * \internal
 * \brief Structure description for HRSStruct
 * \see HRSStruct
 */
struct structFormat *HRSStructStruct;

/**
 * \internal
 * \brief Structure description for FDSStruct
 * \see FecDataStruct
 */
struct structFormat *FDSStructStruct;

/**
 * \internal
 * \brief Structure description for the central logger entry structure: CLOGStruct
 * \see CLOG
 */
struct structFormat *CLOGStructStruct;
/**
 * \internal
 * \brief Initializes and registers pre-defined tagged structures.
 * 
 * \see HeaderStructStruct 
 * \see SUBSCRIPTIONStruct
 * \see ProducerHeaderStructStruct
 * \see ServerStateStructStruct
 * \see RPCDataStructStruct
 * \see RPCFecStructStruct
 * \see ServerStateStructStruct
 * \see GlobalDataHdrStruct
 * \see SpectrumHeaderStruct
 * \see PropertyQueryStruct
 * \see PropertyQueryStructEx
 * \see XPropertyQueryStruct
 * \see ActivityQueryStruct
 * \see ContractQueryStruct
 * \see ClientQueryStruct
 * \see AMSStructStruct
 * \see ADSStructStruct
 */
int initControlStructs(void)
{
  struct structStruct *ss = NULL;
  int cc;

  if (gControlStructInitialized) return 0;
  gControlStructInitialized = TRUE;

# ifndef DOS
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,TotalSize),
		       1,CF_SHORT,NULL,"totalSize")) != 0) return cc;
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,UserName),
		       USERNAME_SIZE,CF_TEXT,NULL,"userName")) != 0) return cc;
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,Protocol),
		       1,CF_SHORT,NULL,"payloadType")) != 0) return cc;
  if ((cc=addtostruct(&HeaderStructStruct,
		       OFFSETIN(HeaderStruct,EqmSize),
		       1,CF_SHORT,NULL,"numParts")) != 0) return cc;

  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,msgsize),
		       1,CF_SHORT,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,extsize),
		       1,CF_SHORT,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,mtu),
           1,CF_SHORT,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,numblks),
		       1,CF_SHORT,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,blknum),
		       1,CF_SHORT,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,id),
		       1,CF_SHORT,NULL,"id")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,mode),
		       1,CF_SHORT,NULL,"mode")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,blkid),
		       1,CF_SHORT,NULL,"blkId")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,PollingRate),
		       1,CF_LONG,NULL,"pollingRate")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,starttime),
		       1,CF_LONG,NULL,"startTime")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmProperty),
		       PROPERTY_NAME_SIZE,CF_TEXT,NULL,"property")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmDeviceName),
		       DEVICE_NAME_SIZE,CF_TEXT,NULL,"deviceName")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmName),
		       EQM_NAME_SIZE,CF_TEXT,NULL,"eqm")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmSizeIn),
		       1,CF_LONG,NULL,"sizeIn")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmSizeOut),
		       1,CF_LONG,NULL,"sizeOut")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.hEqmName),
		       1,CF_BYTE,NULL,"handle")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmAccess),
		       1,CF_BYTE,NULL,"access")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmFormatIn),
		       1,CF_BYTE,NULL,"formatIn")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.EqmFormatOut),
		       1,CF_BYTE,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.strTagIn),
		       TAG_NAME_SIZE,CF_TEXT,NULL,"tagIn")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTIONStruct,
		       OFFSETIN(SUBSCRIPTION,contract.strTagOut),
		       TAG_NAME_SIZE,CF_TEXT,NULL,"tagOut")) != 0) return cc;

  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,msgsize),
		       1,CF_SHORT,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,mtu),
           1,CF_SHORT,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,numblks),
		       1,CF_BYTE,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,blknum),
		       1,CF_BYTE,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,id),
		       1,CF_SHORT,NULL,"id")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,mode),
		       1,CF_SHORT,NULL,"mode")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,blkid),
		       1,CF_SHORT,NULL,"blkId")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,PollingRate),
		       1,CF_LONG,NULL,"pollingRate")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,starttime),
		       1,CF_LONG,NULL,"startTime")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmProperty),
		       PROPERTY_NAME_SHORTSIZE,CF_TEXT,NULL,"property")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmDeviceName),
		       DEVICE_NAME_SHORTSIZE,CF_TEXT,NULL,"deviceName")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmName),
		       EQM_NAME_SIZE,CF_TEXT,NULL,"eqm")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmSizeIn),
		       1,CF_LONG,NULL,"sizeIn")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmSizeOut),
		       1,CF_LONG,NULL,"sizeOut")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.hEqmName),
		       1,CF_BYTE,NULL,"handle")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmAccess),
		       1,CF_BYTE,NULL,"access")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmFormatIn),
		       1,CF_BYTE,NULL,"formatIn")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.EqmFormatOut),
		       1,CF_BYTE,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.strTagIn),
		       TAG_NAME_SHORTSIZE,CF_TEXT,NULL,"tagIn")) != 0) return cc;
  if ((cc=addtostruct(&SUBSCRIPTION5Struct,
		       OFFSETIN(SUBSCRIPTION5,contract.strTagOut),
		       TAG_NAME_SHORTSIZE,CF_TEXT,NULL,"tagOut")) != 0) return cc;

  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,msgsize),
		       1,CF_SHORT,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,SubscriptionID),
		       1,CF_SHORT,NULL,"subId")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,CompletionCode),
		       1,CF_SHORT,NULL,"compCode")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,numblks),
		       1,CF_SHORT,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,blknum),
		       1,CF_SHORT,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,mtu),
           1,CF_SHORT,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,timestampMSEC),
		       1,CF_SHORT,NULL,"timestampMsec")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,EqmFormat),
		       1,CF_BYTE,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,counter),
		       1,CF_BYTE,NULL,"counter")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,timestamp),
		       1,CF_LONG,NULL,"timestampSec")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStruct5Struct,
		       OFFSETIN(ProducerHeaderStruct5,ClnStarttime),
		       1,CF_LONG,NULL,"startTimeTag")) != 0) return cc;

  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,msgsize),
		       1,CF_SHORT,NULL,"msgSize")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,SubscriptionID),
		       1,CF_SHORT,NULL,"subId")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,CompletionCode),
		       1,CF_SHORT,NULL,"compCode")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,numblks),
		       1,CF_SHORT,NULL,"numBlks")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,blknum),
		       1,CF_SHORT,NULL,"blkNum")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,mtu),
           1,CF_SHORT,NULL,"mtu")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,EqmFormat),
		       1,CF_SHORT,NULL,"formatOut")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,counter),
		       1,CF_SHORT,NULL,"counter")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,SrvProtocol),
		       1,CF_SHORT,NULL,"srvProtocol")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,xferReason),
		       1,CF_SHORT,NULL,"xferReason")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,ClnStarttime),
		       1,CF_LONG,NULL,"startTimeTag")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,stssize),
		       1,CF_SHORT,NULL,"statusSize")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,stscode),
		       1,CF_SHORT,NULL,"statusCode")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,timestamp),
		       1,CF_LONG,NULL,"tv_sec")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,timestampUSec),
		       1,CF_LONG,NULL,"tv_usec")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,userstamp),
		       1,CF_LONG,NULL,"userStamp")) != 0) return cc;
  if ((cc=addtostruct(&ProducerHeaderStructStruct,
		       OFFSETIN(ProducerHeaderStruct,sysstamp),
		       1,CF_LONG,NULL,"sysStamp")) != 0) return cc;

  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,FecName),
		       FEC_NAME_SIZE,CF_TEXT,NULL,"fecName")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,SubSystem),
		       5,CF_TEXT,NULL,"subSystem")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,ExportName),
		       EXPORT_NAME_SHORTSIZE,CF_TEXT,NULL,"exportName")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,EqmName),
		       EQM_NAME_SHORTSIZE+1,CF_TEXT,NULL,"eqmName")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,EqmContext),
		       CONTEXT_NAME_SHORTSIZE,CF_TEXT,NULL,"context")) != 0) return cc;
  if ((cc=addtostruct(&RPCDataStructStruct,
		       OFFSETIN(RPCDataStruct,alarmSystem),
		       1,CF_SHORT,NULL,"almSystem")) != 0) return cc;

  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Name),
		       FEC_NAME_SIZE,CF_TEXT,NULL,"fecName")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,reserved),
		       5,CF_BYTE,NULL,"reserved")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Network),
		       4,CF_BYTE,NULL,"ipxNet")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Node),
		       6,CF_BYTE,NULL,"ipxNode")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,ImmediateAddress),
		       6,CF_BYTE,NULL,"ipxImmAddr")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,IP),
		       ADDR_SIZE,CF_TEXT,NULL,"ipAddr")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,PortOffset),
		       1,CF_BYTE,NULL,"portOffset")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,IPh_addr),
		       4,CF_BYTE,NULL,"ip_haddr")) != 0) return cc;
  if ((cc=addtostruct(&RPCFecStructStruct,
		       OFFSETIN(RPCFecStruct,Prot),
		       2,CF_BYTE,NULL,"transportType")) != 0) return cc;

  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,desc),
		       64,CF_TEXT,NULL,"desc")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,dataFmt),
		       1,CF_BYTE,NULL,"dataFmt")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,dataSiz),
		       1,CF_BYTE,NULL,"dataSiz")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,data),
		       64,CF_BYTE,NULL,"data")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,timeEntered),
		       1,CF_LONG,NULL,"timeEntered")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,codeNow),
		       1,CF_LONG,NULL,"codeNow")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,codePrv),
		       1,CF_LONG,NULL,"codePrv")) != 0) return cc;
  if ((cc=addtostruct(&ServerStateStructStruct,
		       OFFSETIN(ServerStateStruct,codeNxt),
		       1,CF_LONG,NULL,"codeNxt")) != 0) return cc;

  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,name),
		       16,CF_TEXT,NULL,"keyword")) != 0) return cc;
  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,size),
		       1,CF_SHORT,NULL,"size")) != 0) return cc;
  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,format),
		       1,CF_SHORT,NULL,"format")) != 0) return cc;
  if ((cc=addtostruct(&GlobalDataHdrStruct,
		       OFFSETIN(GlobalDataHdr,time),
		       1,CF_LONG,NULL,"utc")) != 0) return cc;

  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,comment),
		       80,CF_TEXT,NULL,"comment")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,tm),
		       1,CF_LONG,NULL,"tm")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,s_start),
		       1,CF_FLOAT,NULL,"start")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,s_inc),
		       1,CF_FLOAT,NULL,"inc")) != 0) return cc;
  if ((cc=addtostruct(&SpectrumHeaderStruct,
		       OFFSETIN(hdrSPECTRUM,status),
		       1,CF_LONG,NULL,"status")) != 0) return cc;
# endif

  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,Name),
		       FEC_NAME_SIZE,CF_TEXT,"Name")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,IPXNetwork),
		       4,CF_BYTE,"IPXNetwork")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,IPXNode),
		       6,CF_BYTE,"IPXNode")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,IPXImmediateAddress),
		       6,CF_BYTE,"IPXImmAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,IP),
		       ADDR_SIZE,CF_TEXT,"IP")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,IPh_addr),
		       4,CF_TEXT,"IPh_addr")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,PortOffset),
		       1,CF_LONG,"PortOffset")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,TransportProtocol),
		       1,CF_LONG,"TransProt")) != 0) return cc;
  if ((cc=addFieldToStruct("FDSr4",
		       OFFSETIN(FecDataStruct,TineProtocol),
		       1,CF_LONG,"TineProtocol")) != 0) return cc;
  if ((cc=sealTaggedStruct("FDSr4",sizeof(HRS),1)) != 0) return cc;

  if ((ss=findstruct("FDSr4")) == NULL) return out_of_server_memory;
  FDSStructStruct = ss->s;

  if ((cc=addFieldToStruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpName),
		       PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpDescription),
		       32,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpSize),
		       1,CF_SHORT,"size")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpFormat),
		       1,CF_BYTE,"format")) != 0) return cc;
  if ((cc=addFieldToStruct("PQS",
		       OFFSETIN(PropertyQueryStruct,prpAccess),
		       1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=sealTaggedStruct("PQS",sizeof(PropertyQueryStruct),100)) != 0) return cc;

  if ((ss=findstruct("PQS")) == NULL) return out_of_server_memory;
  PropertyQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpName),
		       PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpDescription),
		       64,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpRedirection),
		       32,CF_TEXT,"redirection")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpTag),
		       TAG_NAME_SHORTSIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpTagIn),
		       TAG_NAME_SHORTSIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpUnits),
		       UNITS_SIZE,CF_TEXT,"units")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpMinValue),
		       1,CF_FLOAT,"min")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpMaxValue),
		       1,CF_FLOAT,"max")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpSize),
		       1,CF_LONG,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpSizeIn),
		       1,CF_LONG,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpNumOverloads),
		       1,CF_LONG,"overloads")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpHistoryDepthShort),
		       1,CF_SHORT,"historyLong")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpHistoryDepthShort),
		       1,CF_SHORT,"historyShort")) != 0) return cc;
  if ((cc=addFieldToStruct("PQSX",
		       OFFSETIN(PropertyQueryStructEx,prpFormat),
		       4,CF_BYTE,"formats")) != 0) return cc;
  if ((cc=sealTaggedStruct("PQSX",sizeof(PropertyQueryStructEx),100)) != 0) return cc;

  if ((ss=findstruct("PQSX")) == NULL) return out_of_server_memory;
  PropertyQueryStructExStruct = ss->s;

  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpName),
		       PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpDescription),
		       64,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpRedirection),
		       32,CF_TEXT,"redirection")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpTag),
		       TAG_NAME_SHORTSIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpTagIn),
		       TAG_NAME_SHORTSIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpUnits),
		       UNITS_SIZE,CF_TEXT,"units")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpMinValue),
		       1,CF_FLOAT,"min")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpMaxValue),
		       1,CF_FLOAT,"max")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpSize),
		       1,CF_LONG,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpSizeIn),
		       1,CF_LONG,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpNumOverloads),
		       1,CF_LONG,"overloads")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpHistoryDepthShort),
		       1,CF_SHORT,"historyShort")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpHistoryDepthLong),
		       1,CF_SHORT,"historyLong")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpFormat),
		       1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpFormatIn),
		       1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpAccess),
		       1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpGraphType),
		       1,CF_BYTE,"graphType")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,rngUnits),
		       UNITS_SIZE,CF_TEXT,"rangeUnits")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,rngMinValue),
		       1,CF_FLOAT,"rangeMin")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,rngMaxValue),
		       1,CF_FLOAT,"rangeMax")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,numRows),
		       1,CF_SHORT,"numRows")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,rowSize),
		       1,CF_SHORT,"rowSize")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,prpArrayType),
		       1,CF_SHORT,"arrayType")) != 0) return cc;
  if ((cc=addFieldToStruct("XPQS",
		       OFFSETIN(XPropertyQueryStruct,reserved),
		       3,CF_SHORT,"reserved")) != 0) return cc;
  if ((cc=sealTaggedStruct("XPQS",sizeof(XPropertyQueryStruct),10)) != 0) return cc;

  if ((ss=findstruct("XPQS")) == NULL) return out_of_server_memory;
  XPropertyQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpName),
		       PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpDescription),
		       PROPERTY_DESC_SIZE,CF_TEXT,"description")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpRedirection),
		       PROPERTY_REDIR_SIZE,CF_TEXT,"redirection")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpTag),
		       TAG_NAME_SIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpTagIn),
		       TAG_NAME_SIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpUnits),
		       UNITS_SIZE,CF_TEXT,"units")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpMinValue),
		       1,CF_FLOAT,"min")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpMaxValue),
		       1,CF_FLOAT,"max")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpSize),
		       1,CF_LONG,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpSizeIn),
		       1,CF_LONG,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpNumOverloads),
		       1,CF_LONG,"overloads")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpHistoryDepthShort),
		       1,CF_SHORT,"historyShort")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpHistoryDepthLong),
		       1,CF_SHORT,"historyLong")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpFormat),
		       1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpFormatIn),
		       1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpAccess),
		       1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpGraphType),
		       1,CF_BYTE,"graphType")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,rngUnits),
		       UNITS_SIZE,CF_TEXT,"rangeUnits")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,rngMinValue),
		       1,CF_FLOAT,"rangeMin")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,rngMaxValue),
		       1,CF_FLOAT,"rangeMax")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,numRows),
		       1,CF_SHORT,"numRows")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,rowSize),
		       1,CF_SHORT,"rowSize")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,prpArrayType),
		       1,CF_SHORT,"arrayType")) != 0) return cc;
  if ((cc=addFieldToStruct("PRPQSr4",
		       OFFSETIN(PrpQueryStruct,reserved),
		       3,CF_SHORT,"reserved")) != 0) return cc;
  if ((cc=sealTaggedStruct("PRPQSr4",sizeof(PrpQueryStruct),10)) != 0) return cc;

  if ((ss=findstruct("PRPQSr4")) == NULL) return out_of_server_memory;
  PrpQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,FecName),
		       FEC_NAME_SIZE,CF_TEXT,"fecName")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,reserved),
		       4,CF_TEXT,"reserved")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,localtime),
		       1,CF_LONG,"localtime")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,starttime),
		       1,CF_LONG,"starttime")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,systemPollingRate),
		       1,CF_LONG,"sysPollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numBkgTsks),
		       1,CF_SHORT,"numBkgTasks")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTotalContracts),
		       1,CF_SHORT,"numContracts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTotalClients),
		       1,CF_SHORT,"numClients")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTargetContracts),
		       1,CF_SHORT,"numTgtCont")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTargetClients),
		       1,CF_SHORT,"numTgtCln")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numConnections),
		       1,CF_SHORT,"numConnections")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numConnectionTimeouts),
		       1,CF_LONG,"numTimeouts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numConnectionArrivals),
		       1,CF_LONG,"numArrivals")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numUDPpkts),
		       1,CF_LONG,"numUDPPkts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numTCPpkts),
		       1,CF_LONG,"numTCPPkts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numIPXpkts),
		       1,CF_LONG,"numIPXPkts")) != 0) return cc;
  if ((cc=addFieldToStruct("AQS",
		       OFFSETIN(ActivityQueryStruct,numSPXpkts),
		       1,CF_LONG,"numSPSPkts")) != 0) return cc;
  if ((cc=sealTaggedStruct("AQS",sizeof(ActivityQueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("AQS")) == NULL) return out_of_server_memory;
  ActivityQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmProperty),
		       PROPERTY_NAME_SHORTSIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmDeviceName),
		       DEVICE_NAME_SHORTSIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmName),
		       EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmSizeIn),
		       1,CF_LONG,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmSizeOut),
		       1,CF_LONG,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.hEqmName),
		       1,CF_BYTE,"hEqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmAccess),
		       1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmFormatIn),
		       1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.EqmFormatOut),
		       1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.strTagIn),
		       TAG_NAME_SHORTSIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,contract.strTagOut),
		       TAG_NAME_SHORTSIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,pollrate),
		       1,CF_LONG,"pollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("CONQS",
		       OFFSETIN(Contract5QueryStruct,nconsumers),
		       1,CF_LONG,"nClients")) != 0) return cc;
  if ((cc=sealTaggedStruct("CONQS",sizeof(Contract5QueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("CONQS")) == NULL) return out_of_server_memory;
  CONTRACT5Struct = ss->s;

  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmProperty),
		       PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmDeviceName),
		       DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmName),
		       EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmSizeIn),
		       1,CF_LONG,"sizeIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmSizeOut),
		       1,CF_LONG,"sizeOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.hEqmName),
		       1,CF_BYTE,"hEqm")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmAccess),
		       1,CF_BYTE,"access")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmFormatIn),
		       1,CF_BYTE,"formatIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.EqmFormatOut),
		       1,CF_BYTE,"formatOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.strTagIn),
		       TAG_NAME_SIZE,CF_TEXT,"tagIn")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,contract.strTagOut),
		       TAG_NAME_SIZE,CF_TEXT,"tagOut")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,pollrate),
		       1,CF_LONG,"pollRate")) != 0) return cc;
  if ((cc=addFieldToStruct("CTQSr4",
		       OFFSETIN(ContractQueryStruct,nconsumers),
		       1,CF_LONG,"nClients")) != 0) return cc;
  if ((cc=sealTaggedStruct("CTQSr4",sizeof(ContractQueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("CTQSr4")) == NULL) return out_of_server_memory;
  CONTRACTStruct = ss->s;

  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.userName),
		       USERNAME_SIZE,CF_TEXT,"username")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPXaddress),
		       12,CF_BYTE,"ipxAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.node),
		       6,CF_BYTE,"ipxNode")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_family),
		       1,CF_SHORT,"ipFamily")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_port),
		       1,CF_SHORT,"ipPort")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_addr.s_addr),
		       4,CF_BYTE,"ipAddr")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.IPaddress.sin_zero),
		       8,CF_BYTE,"ipZero")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.count),
		       1,CF_SHORT,"numContracts")) != 0) return cc;
  if ((cc=addFieldToStruct("CLNQS",
		       OFFSETIN(ClientQueryStruct,cln.protocol),
		       1,CF_SHORT,"transport")) != 0) return cc;
  if ((cc=sealTaggedStruct("CLNQS",sizeof(ClientQueryStruct),1)) != 0) return cc;

  if ((ss=findstruct("CLNQS")) == NULL) return out_of_server_memory;
  ClientQueryStructStruct = ss->s;

  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,server),
		       EXPORT_NAME_SHORTSIZE,CF_TEXT,"server")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,device),
		       DEVICE_NAME_SHORTSIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmTag),
		       ALARM_TAG_SHORTSIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmCode),
		       1,CF_LONG,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,timestamp),
		       1,CF_LONG,"almTime")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmMask),
		       1,CF_LONG,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmData),
		       ALARM_DATA_SHORTSIZE,CF_BYTE,"almData")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmDataFormat),
		       1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmDataArraySize),
		       1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,severity),
		       1,CF_BYTE,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,descriptor),
		       1,CF_BYTE,"almDescriptor")) != 0) return cc;
  if ((cc=addFieldToStruct("AMS",
		       OFFSETIN(AMS5,alarmSystem),
		       1,CF_SHORT,"almSystem")) != 0) return cc;
  if ((cc=sealTaggedStruct("AMS",sizeof(AMS5),1000)) != 0) return cc;

  if ((ss=findstruct("AMS")) == NULL) return out_of_server_memory;
  AMS5StructStruct = ss->s;
    
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,server),
		       EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,device),
		       DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmTag),
		       ALARM_TAG_SIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmCode),
		       1,CF_LONG,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,timestamp),
		       1,CF_LONG,"almTime")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,timestampUSec),
		       1,CF_LONG,"almTimeUSec")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,starttime),
		       1,CF_LONG,"almStart")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,starttimeUSec),
		       1,CF_LONG,"almStartUSec")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmMask),
		       1,CF_LONG,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmData),
		       ALARM_DATA_SIZE,CF_BYTE,"almData")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmDataFormat),
		       1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmDataArraySize),
		       1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,severity),
		       1,CF_BYTE,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,descriptor),
		       1,CF_BYTE,"almDescriptor")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,alarmSystem),
		       1,CF_SHORT,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("AMSr4",
		       OFFSETIN(AMS,reserved),
		       2,CF_BYTE,"reserved")) != 0) return cc;
  if ((cc=sealTaggedStruct("AMSr4",sizeof(AMS),1000)) != 0) return cc;

  if ((ss=findstruct("AMSr4")) == NULL) return out_of_server_memory;
  AMSStructStruct = ss->s;

  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmTag),
		       ALARM_TAG_SHORTSIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmCode),
		       1,CF_LONG,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmMask),
		       1,CF_LONG,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,severity),
		       1,CF_SHORT,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmDataFormat),
		       1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmDataArraySize),
		       1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,timeRes),
		       1,CF_LONG,"timeRes")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,hitsChannel),
		       1,CF_SHORT,"timeChan")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,hitsModule),
		       1,CF_SHORT,"timeModule")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmText),
		       ALARM_TEXT_SHORTSIZE,CF_TEXT,"almText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,deviceText),
		       ALARM_TEXT_SHORTSIZE-2,CF_TEXT,"devText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,alarmSystem),
		       1,CF_SHORT,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,dataText),
		       ALARM_TEXT_SHORTSIZE,CF_TEXT,"dataText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,url),
		       ALARM_TEXT_SHORTSIZE,CF_TEXT,"url")) != 0) return cc;
  if ((cc=addFieldToStruct("ADS",
		       OFFSETIN(ADS5,ndefs),
		       1,CF_LONG,"numDefs")) != 0) return cc;
  if ((cc=sealTaggedStruct("ADS",sizeof(ADS5),100)) != 0) return cc;

  if ((ss=findstruct("ADS")) == NULL) return out_of_server_memory;
  ADS5StructStruct = ss->s;

  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmTag),
		       ALARM_TAG_SIZE,CF_TEXT,"almTag")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmCode),
		       1,CF_LONG,"almCode")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmMask),
		       1,CF_LONG,"almMask")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmSystem),
		       1,CF_SHORT,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmSeverity),
		       1,CF_SHORT,"almSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmDataFormat),
		       1,CF_BYTE,"almDataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmDataArraySize),
		       1,CF_BYTE,"almDataLen")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,reserved),
		       2,CF_BYTE,"reserved")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,alarmText),
		       ALARM_TEXT_SIZE,CF_TEXT,"almText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,deviceText),
		       ALARM_TEXT_SIZE,CF_TEXT,"devText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,dataText),
		       ALARM_TEXT_SIZE,CF_TEXT,"dataText")) != 0) return cc;
  if ((cc=addFieldToStruct("ADSr4",
		       OFFSETIN(ADS,url),
		       ALARM_TEXT_LONGSIZE,CF_TEXT,"url")) != 0) return cc;
  if ((cc=sealTaggedStruct("ADSr4",sizeof(ADS),100)) != 0) return cc;

  if ((ss=findstruct("ADSr4")) == NULL) return out_of_server_memory;
  ADSStructStruct = ss->s;

  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,eqm),
		       EQM_NAME_SIZE,CF_TEXT,"eqm")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,dev),
		       DEVICE_NAME_SIZE,CF_TEXT,"dev")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,prp),
		       PROPERTY_NAME_SIZE,CF_TEXT,"prp")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,siz),
		       1,CF_LONG,"len")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,fmt),
		       1,CF_LONG,"fmt")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,atyp),
		       1,CF_LONG,"arrayType")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,asys),
		       1,CF_LONG,"almSystem")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,sev),
		       1,CF_LONG,"genSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,hisev),
		       1,CF_LONG,"hiSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,hiwarnsev),
		       1,CF_LONG,"hiwarnSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,losev),
		       1,CF_LONG,"loSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,lowarnsev),
		       1,CF_LONG,"lowarnSeverity")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,chkNormal),
		       1,CF_LONG,"valueCheck")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,mask),
		       1,CF_LONG,"valueMask")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,normal),
		       1,CF_LONG,"valueNormal")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,cnt),
		       1,CF_LONG,"cnts")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,cntThreshold),
		       1,CF_LONG,"cntsThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,hi),
		       1,CF_FLOAT,"hiThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,lo),
		       1,CF_FLOAT,"loThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,hiwarn),
		       1,CF_FLOAT,"hiwarnThreshold")) != 0) return cc;
  if ((cc=addFieldToStruct("AWSr4",
		       OFFSETIN(AWS,lo),
		       1,CF_FLOAT,"lowarnThreshold")) != 0) return cc;
  if ((cc=sealTaggedStruct("AWSr4",sizeof(AWS),100)) != 0) return cc;

  if ((ss=findstruct("AWSr4")) == NULL) return out_of_server_memory;
  AWSStructStruct = ss->s;

  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,property),
		       PROPERTY_NAME_SIZE,CF_TEXT,"property")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,device),
		       DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,datasiz),
		       1,CF_LONG,"dataSize")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,datafmt),
		       1,CF_LONG,"dataFormat")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,pollingRate),
		       1,CF_LONG,"pollingRate")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,archiveRate),
		       1,CF_LONG,"archiveRate")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,depthShort),
		       1,CF_LONG,"depthShort")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,depthLong),
		       1,CF_LONG,"depthLong")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,heartbeat),
		       1,CF_LONG,"heartbeat")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,pTolerance),
		       1,CF_FLOAT,"tolerancePct")) != 0) return cc;
  if ((cc=addFieldToStruct("HRSr4",
		       OFFSETIN(HRS,aTolerance),
		       1,CF_FLOAT,"toleranceAbs")) != 0) return cc;
  if ((cc=sealTaggedStruct("HRSr4",sizeof(HRS),1)) != 0) return cc;

  if ((ss=findstruct("HRSr4")) == NULL) return out_of_server_memory;
  HRSStructStruct = ss->s;

  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,context),
		       32,CF_TEXT,"context")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,tag),
		       32,CF_TEXT,"tag")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,logger),
		       32,CF_TEXT,"logger")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,text),
		       128,CF_TEXT,"text")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,priority),
		       1,CF_LONG,"priority")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,status),
		       1,CF_LONG,"status")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,timeSent),
		       1,CF_DOUBLE,"timeSent")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOG",
		       OFFSETIN(CLOG,timeLogged),
		       1,CF_DOUBLE,"timeLogged")) != 0) return cc;
  if ((cc=sealTaggedStruct("CLOG",sizeof(CLOG),1000)) != 0) return cc;

  if ((ss=findstruct("CLOG")) == NULL) return out_of_server_memory;
  CLOGStructStruct = ss->s;

  feclog("Control Structures Swap information registered");
  return 0;
}
